"""
gpu.py: general-purpose utilities for GPU
"""
from ..utils import check_supported

try:
    from pycuda.driver import Device as CudaDevice
    from pycuda.driver import device_attribute as dev_attrs

    __has_pycuda__ = True
except ImportError:
    __has_pycuda__ = False
    CudaDevice = type(None)
try:
    from pyopencl import Device as CLDevice

    __has_pyopencl__ = True
except ImportError:
    CLDevice = type(None)
    __has_pyopencl__ = False

#
# silx.opencl.common.Device cannot be supported as long as
# silx.opencl instantiates the "ocl" singleton in __init__,
# leaving opencl contexts all over the place in some cases
#


class GPUDescription:
    """
    Simple description of a Graphical Processing Unit.
    This class is designed to be simple to understand, and to be serializable
    for being used by dask.distributed.
    """

    def __init__(self, device, vendor=None, device_id=None):
        """
        Create a description from a device.

        Parameters
        ----------
        device: `pycuda.driver.Device` or `pyopencl.Device`
            Class describing a GPU device.
        """
        is_cuda_device = isinstance(device, CudaDevice)
        is_cl_device = isinstance(device, CLDevice)

        if is_cuda_device:
            self._init_from_cuda_device(device)
        elif is_cl_device:
            self._init_from_cl_device(device)
            self._set_other_attrs(vendor, device_id)
        else:
            raise ValueError("Expected `pycuda.driver.Device` or `pyopencl.Device`")

    def _init_from_cuda_device(self, device):
        self._dict = {
            "type": "cuda",
            "name": device.name(),
            "memory_GB": device.total_memory() / 1e9,
            "compute_capability": device.compute_capability(),
            "device_id": device.get_attribute(dev_attrs.MULTI_GPU_BOARD_GROUP_ID),
        }

    def _init_from_cl_device(self, device):
        self._dict = {
            "type": "opencl",
            "name": device.name,
            "memory_GB": device.global_mem_size / 1e9,
            "vendor": device.vendor,
        }

    def _set_other_attrs(self, vendor, device_id):
        if vendor is not None:
            self._dict["vendor"] = vendor
        if device_id is not None:
            self._dict["device_id"] = device_id  # device ID for OpenCL (!= platform ID)

    def _dict_to_self(self):
        for key, val in self._dict.items():
            setattr(self, key, val)

    def get_dict(self):
        return self._dict


GPU_PICK_METHODS = ["cuda", "auto"]


def pick_gpus(method, cuda_gpus, opencl_platforms, n_gpus):
    check_supported(method, GPU_PICK_METHODS, "GPU picking method")
    if method == "cuda":
        return pick_gpus_nvidia(cuda_gpus, n_gpus)
    elif method == "auto":
        return pick_gpus_auto(cuda_gpus, opencl_platforms, n_gpus)
    else:
        return []


# TODO Fix this function, it is broken:
#  - returns something when n_gpus = 0
#  - POCL increments device_id by 1 !
def pick_gpus_auto(cuda_gpus, opencl_platforms, n_gpus):
    """
    Pick `n_gpus` devices with the best available driver.

    This function browse the visible Cuda GPUs and Opencl platforms to pick
    the GPUs with the best driver.
    A worker might see several implementations of a GPU driver.
    For example with Nvidia hardware, we can see:
      - The Cuda implementation (nvidia-cuda-toolkit)
      - OpenCL implementation by Nvidia (nvidia-opencl-icd)
      - OpenCL implementation by Portable OpenCL

    Parameters
    ----------
    cuda_gpu: dict
        Dictionary where each key is an ID, and the value is a dictionary
        describing some attributes of the GPU (obtained with `GPUDescription`)
    opencl_platforms: dict
        Dictionary where each key is the platform name, and the value is a list
        of dictionary descriptions.
    n_gpus: int
        Number of GPUs to pick.
    """

    def gpu_equal(gpu1, gpu2):
        # TODO find a better test ?
        # Some information are not always available depending on the opencl vendor !
        return (gpu1["device_id"] == gpu2["device_id"]) and (gpu1["name"] == gpu2["name"])

    def is_in_gpus(avail_gpus, query_gpu):
        for gpu in avail_gpus:
            if gpu_equal(gpu, query_gpu):
                return True
        return False

    # If some Nvidia hardware is visible, add it without question.
    # In the case we don't want it, we should either re-run resources discovery
    # with `try_cuda=False`, or mask individual devices with CUDA_VISIBLE_DEVICES.
    chosen_gpus = list(cuda_gpus.values())
    if len(chosen_gpus) >= n_gpus:
        return chosen_gpus
    for platform, gpus in opencl_platforms.items():
        for gpu_id, gpu in gpus.items():
            if not (is_in_gpus(chosen_gpus, gpu)):
                # TODO prioritize some OpenCL implementations ?
                chosen_gpus.append(gpu)
    if len(chosen_gpus) < n_gpus:
        raise ValueError("Not enough GPUs: could only collect %d/%d" % (len(chosen_gpus), n_gpus))
    return chosen_gpus


def pick_gpus_nvidia(cuda_gpus, n_gpus):
    """
    Pick one or more Nvidia GPUs.
    """
    if len(cuda_gpus) < n_gpus:
        raise ValueError("Not enough Nvidia GPU: requested %d, but can get only %d" % (n_gpus, len(cuda_gpus)))
    # Sort GPUs by computing capabilities, pick the "best" ones
    gpus_cc = []
    for gpu_id, gpu in cuda_gpus.items():
        cc = gpu["compute_capability"]
        gpus_cc.append((gpu_id, cc[0] + 0.1 * cc[1]))
    gpus_cc_sorted = sorted(gpus_cc, key=lambda x: x[1], reverse=True)
    res = []
    for i in range(n_gpus):
        res.append(cuda_gpus[gpus_cc_sorted[i][0]])
    return res



from .cli_configs import BootstrapStitchingConfig
from ..pipeline.config import generate_nabu_configfile
from ..stitching.config import get_default_stitching_config, SECTIONS_COMMENTS as _SECTIONS_COMMENTS
from .utils import parse_params_values


def bootstrap_stitching():
    args = parse_params_values(
        BootstrapStitchingConfig,
        parser_description="Initialize a nabu configuration file",
    )

    prefilled_values = {}

    generate_nabu_configfile(
        fname=args["output"],
        default_config=get_default_stitching_config(args["stitching_type"]),
        comments=True,
        sections_comments=_SECTIONS_COMMENTS,
        options_level=args["level"],
        prefilled_values=prefilled_values,
    )



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "10/05/2022"


from copy import copy
from datetime import datetime
from multiprocessing.sharedctypes import Value
import os
from silx.io.utils import get_data
from typing import Optional, Union
from silx.io.url import DataUrl
from silx.io.dictdump import dicttonx
import numpy
from tomoscan.esrf import HDF5TomoScan, EDFTomoScan
from tomoscan.serie import Serie
from tomoscan.esrf.scan.hdf5scan import ImageKey
from tomoscan.nexus.paths.nxtomo import get_paths as _get_nexus_paths
from tomoscan.scanbase import TomoScanBase
from scipy.ndimage import shift as shift_scipy
from nxtomomill.nexus import NXtomo
from nabu.io.utils import DatasetReader
from tomoscan.esrf.scan.utils import (
    get_compacted_dataslices,
)  # this version has a 'return_url_set' needed here. At one point they should be merged together
from nabu.stitching.frame_composition import ZFrameComposition
from nabu.stitching.utils import find_relative_shifts
from nabu.stitching.config import (
    KEY_CROSS_CORRELATION_SLICE,
    KEY_X_CROSS_CORRELATION_FUNC,
    KEY_Y_CROSS_CORRELATION_FUNC,
    PreProcessedZStitchingConfiguration,
    PostProcessedZStitchingConfiguration,
    ZStitchingConfiguration,
)
from nabu.utils import Progress
from nabu import version as nabu_version
from nabu.io.writer import get_datetime
from .overlap import (
    DEFAULT_OVERLAP_STRATEGY,
    OverlapKernelBase,
    OverlapStichingStrategy,
    ZStichOverlapKernel,
)
from tomoscan.io import HDF5File
import h5py
import logging

_logger = logging.getLogger(__name__)


def z_stitching(configuration: ZStitchingConfiguration, progress=None) -> DataUrl:
    """
    Apply stitching from provided configuration.
    Return a DataUrl with the created NXtomo or Volume
    """
    if isinstance(configuration, PreProcessedZStitchingConfiguration):
        stitcher = PreProcessZStitcher(configuration=configuration, progress=progress)
    elif isinstance(configuration, PostProcessedZStitchingConfiguration):
        raise NotImplementedError
    else:
        raise TypeError(
            f"configuration is expected to be in {(PreProcessedZStitchingConfiguration, PostProcessedZStitchingConfiguration)}. {type(configuration)} provided"
        )
    return stitcher.stitch()


class ZStitcher:
    def __init__(self, configuration, progress: Progress = None) -> None:
        if not isinstance(configuration, ZStitchingConfiguration):
            raise TypeError
        self._configuration = copy(configuration)
        # copy configuration because we will edit it
        self._progress = progress
        # z serie must be defined from daughter class
        assert hasattr(self, "_z_serie")

        def is_auto(param):
            return param in ("auto", ("auto",))

        # 'expend' auto shift request if only set once for all
        if is_auto(self.configuration.x_shifts):
            self.configuration.x_shifts = [
                "auto",
            ] * (len(self.z_serie) - 1)
        elif numpy.isscalar(self.configuration.x_shifts):
            self.configuration.x_shifts = [
                self.configuration.x_shifts,
            ] * (len(self.z_serie) - 1)

        # 'expend' overlaph height and
        if is_auto(self.configuration.overlap_height):
            self.configuration.overlap_height = [
                "auto",
            ] * (len(self.z_serie) - 1)
        elif numpy.isscalar(self.configuration.overlap_height):
            self.configuration.overlap_height = [self.configuration.overlap_height] * (len(self.z_serie) - 1)

        # 'expend' stitching height
        if is_auto(self.configuration.stitching_height):
            self.configuration.stitching_height = [
                "auto",
            ] * (len(self.z_serie) - 1)
        elif numpy.isscalar(self._configuration.stitching_height):
            self.configuration.stitching_height = [self.configuration.stitching_height] * (len(self.z_serie) - 1)

    def stitch(self) -> DataUrl:
        """
        Apply expected stitch from configuration and return the DataUrl of the object created
        """
        raise NotImplementedError("base class")

    def _compute_shifts(self):
        raise NotImplementedError("base class")

    @property
    def z_serie(self) -> Serie:
        return self._z_serie

    @property
    def configuration(self) -> ZStitchingConfiguration:
        return self._configuration

    @property
    def progress(self) -> Optional[Progress]:
        return self._progress

    @staticmethod
    def get_overlap_areas(
        lower_frame: numpy.ndarray,
        upper_frame: numpy.ndarray,
        real_overlap: int,
        stitching_height: int,
    ):
        """
        return the requested area from lower_frame and upper_frame.

        Lower_frame contains at the end of it the 'real overlap' with the upper_frame.
        Upper_frame contains the 'real overlap' at the end of it.

        For some reason the user can ask the stitching height to be smaller than the `real overlap`.

        Here are some drawing to have a better of view of those regions:

        .. image:: images/stitching/z_stitch_real_overlap.png
            :width: 600

        .. image:: z_stitch_stitch_height.png
            :width: 600
        """
        assert real_overlap >= 0
        assert stitching_height >= 0
        if stitching_height > real_overlap:
            raise ValueError(f"stitching height ({stitching_height}) larger than existing overlap ({real_overlap}).")
        real_overlap_0 = lower_frame[-real_overlap:]
        real_overlap_1 = upper_frame[:real_overlap]
        if not real_overlap_0.shape == real_overlap_1.shape:
            raise RuntimeError(
                f"lower and upper frame have different overlap size ({real_overlap_0.shape} vs {real_overlap_1.shape})"
            )
        low_pos = int(real_overlap // 2 - stitching_height // 2)
        hight_pos = int(real_overlap // 2 + stitching_height // 2) + (stitching_height) % 2
        # if there is one more line to take on one side take it on the lower_frame side
        assert real_overlap_0[low_pos:hight_pos].shape == real_overlap_1[low_pos:hight_pos].shape
        return real_overlap_0[low_pos:hight_pos], real_overlap_1[low_pos:hight_pos]

    @staticmethod
    def stitch_frames(
        frames: tuple,
        x_relative_shifts: tuple,
        y_overlap_heights: tuple,
        output_dtype: numpy.ndarray,
        output_dataset: Optional[Union[h5py.Dataset, numpy.ndarray]] = None,
        check_inputs=True,
        shift_mode="nearest",
        overlap_kernels=None,
        i_frame=None,
    ) -> numpy.ndarray:
        """
        shift frames according to provided `shifts` (as y, x tuples) then stitch all the shifted frames together and
        save them to output_dataset.

        :param tuple frames: element must be a DataUrl or a 2D numpy array
        """
        if check_inputs:
            if len(frames) < 2:
                raise ValueError(f"Not enought frames provided for stitching ({len(frames)} provided)")
            if len(frames) != len(x_relative_shifts) + 1:
                raise ValueError(
                    f"Incoherent number of shift provided ({len(x_relative_shifts)}) compare to number of frame ({len(frames)}). len(frames) - 1 expected"
                )
            if len(x_relative_shifts) != len(y_overlap_heights):
                raise ValueError(
                    f"expect to have the same number of x_relative_shifts ({len(x_relative_shifts)}) and y_overlap ({len(y_overlap_heights)})"
                )

        def check_frame_is_2d(frame):
            if frame.ndim != 2:
                raise ValueError(f"2D frame expected when {frame.ndim}D provided")

        # step_0 load data if from url
        data = []
        for frame in frames:
            if isinstance(frame, DataUrl):
                data_frame = get_data(frame)
                if check_inputs:
                    check_frame_is_2d(data_frame)
                data.append(data_frame)
            elif isinstance(frame, numpy.ndarray):
                if check_inputs:
                    check_frame_is_2d(frame)
                data.append(frame)
            else:
                raise TypeError(f"frames are expected to be DataUrl or 2D numpy array. Not {type(frame)}")
        # step 1: shift each frames (except the first one)
        x_shifted_data = [data[0]]
        for frame, x_relative_shift in zip(data[1:], x_relative_shifts):
            # note: for now we only shift data in x. the y shift is handled in the FrameComposition
            x_relative_shift = numpy.asarray(x_relative_shift).astype(numpy.int8)
            if x_relative_shift == 0:
                shifted_frame = frame
            else:
                # TO speed up: should use the Fourier transform
                shifted_frame = shift_scipy(
                    frame,
                    mode=shift_mode,
                    shift=[0, -x_relative_shift],
                    order=1,
                )
            x_shifted_data.append(shifted_frame)

        # step 2: create stitched frame
        if overlap_kernels is None:
            overlap_kernels = ZStichOverlapKernel(frame_width=data[0].shape[1])
        stitched_frame = z_stitch_raw_frames(
            frames=x_shifted_data,
            y_shifts=[-abs(y_overlap_height) for y_overlap_height in y_overlap_heights],
            overlap_kernels=overlap_kernels,
            check_inputs=check_inputs,
            output_dtype=output_dtype,
        )

        # step 3: dump stitched frame
        if output_dataset is not None and i_frame is not None:
            output_dataset[i_frame] = stitched_frame
        return stitched_frame


class PreProcessZStitcher(ZStitcher):
    def __init__(self, configuration, progress=None) -> None:
        # z serie must be defined first
        self._z_serie = Serie("z-serie", iterable=configuration.input_scans, use_identifiers=False)
        self._reading_orders = []
        self._x_flips = []
        self._y_flips = []
        # some scan can have been taken in the opposite order (so must be read on the opposite order one from the other)
        super().__init__(configuration, progress)

    @property
    def reading_order(self):
        """
        as scan can be take on one direction or the order (rotation goes from X to Y then from Y to X)
        we might need to read data from one direction or another
        """
        return self._reading_orders

    @property
    def x_flips(self) -> list:
        return self._x_flips

    @property
    def y_flips(self) -> list:
        return self._y_flips

    def stitch(self):
        if self.progress is not None:
            self.progress.set_name("order scans")
        self._order_scans()
        if self.progress is not None:
            self.progress.set_name("check inputs")
        self._check_inputs()
        if self.progress is not None:
            self.progress.set_name("compute flat field")
        self._compute_reduced_flats_and_darks()
        if self.progress is not None:
            self.progress.set_name("compute shift")
        self._compute_shifts()
        if self.progress is not None:
            self.progress.set_name("stitch projections, save them and create NXtomo")
        self._create_nx_tomo()
        if self.progress is not None:
            self.progress.set_name("dump configuration")
        self._dump_stitching_configuration()
        return DataUrl(
            file_path=self.configuration.output_file_path,
            data_path=self.configuration.output_data_path,
            scheme="h5py",
        )

    def _order_scans(self):
        """
        ensure scans are in z increasing order
        """

        def get_min_z(scan):
            return scan.get_bounding_box(axis="z").min

        sorted_z_serie = Serie(
            self.z_serie.name,
            sorted(self.z_serie[:], key=get_min_z, reverse=True),
            use_identifiers=False,
        )
        if sorted_z_serie != self.z_serie:
            if sorted_z_serie[:] != self.z_serie[::-1]:
                raise ValueError("Unable to get comprehensive input. Z (decreasing) ordering is not respected.")
            else:
                _logger.warning(
                    f"z decreasing order haven't been respected. Need to reorder z serie ({[str(scan) for scan in sorted_z_serie[:]]}). Will also reorder overlap height, stitching height and invert shifts"
                )

            self.configuration.overlap_height = self.configuration.overlap_height[::-1]
            self.configuration.x_shifts = [
                -x_shift if x_shift != "auto" else x_shift for x_shift in self.configuration.x_shifts
            ]
            self.configuration.stitching_height = self.configuration.stitching_height[::-1]

        self._z_serie = sorted_z_serie

    def _check_inputs(self):
        """
        insure input data is coherent
        """
        n_scans = len(self.z_serie)
        if n_scans == 0:
            raise ValueError("no scan to stich together")

        # check number of shift provided
        if len(self.configuration.x_shifts) != (n_scans - 1):
            raise ValueError(f"expect {n_scans -1} shift defined. Get {len(self.configuration.x_shifts)}")

        if len(self.configuration.overlap_height) != (n_scans - 1):
            raise ValueError(f"expect {n_scans - 1} overlap defined. Get {len(self.configuration.overlap_height)}")

        if len(self.configuration.stitching_height) != (n_scans - 1):
            raise ValueError(
                f"expect {n_scans - 1} stitching height defined. Get {len(self.configuration.overlap_height)}"
            )

        for scan in self.z_serie:
            if scan.x_flipped is None or scan.y_flipped is None:
                _logger.warning(
                    f"Found at least one scan with no frame flips information ({scan}). Will consider those are unflipped. Might end up with some inverted frame errors."
                )
                break

        self._reading_orders = []
        # the first scan will define the expected reading orderd, and expected flip.
        # if all scan are flipped then we will keep it this way
        self._reading_orders.append(1)

        # check scans are coherent (nb projections, rotation angle, energy...)
        for scan_0, scan_1 in zip(self.z_serie[0:-1], self.z_serie[1:]):
            if len(scan_0.projections) != len(scan_1.projections):
                raise ValueError(f"{scan_0} and {scan_1} have a different number of projections")
            if isinstance(scan_0, HDF5TomoScan) and isinstance(scan_1, HDF5TomoScan):
                # check rotation (only of is an HDF5TomoScan)
                scan_0_angles = numpy.asarray(scan_0.rotation_angle)
                scan_0_projections_angles = scan_0_angles[
                    numpy.asarray(scan_0.image_key_control) == ImageKey.PROJECTION.value
                ]
                scan_1_angles = numpy.asarray(scan_1.rotation_angle)
                scan_1_projections_angles = scan_1_angles[
                    numpy.asarray(scan_1.image_key_control) == ImageKey.PROJECTION.value
                ]
                if not numpy.allclose(scan_0_projections_angles, scan_1_projections_angles, atol=10e-1):
                    if numpy.allclose(
                        scan_0_projections_angles,
                        scan_1_projections_angles[::-1],
                        atol=10e-1,
                    ):
                        reading_order = -1 * self._reading_orders[-1]
                    else:
                        raise ValueError(f"Angles from {scan_0} and {scan_1} are different")
                else:
                    reading_order = 1 * self._reading_orders[-1]
                self._reading_orders.append(reading_order)
            # check energy
            if scan_0.energy is None:
                _logger.warning(f"no energy found for {scan_0}")
            elif not numpy.isclose(scan_0.energy, scan_1.energy, rtol=1e-03):
                _logger.warning(
                    f"different energy found between {scan_0} ({scan_0.energy}) and {scan_1} ({scan_1.energy})"
                )
            # check FOV
            if not scan_0.field_of_view == scan_1.field_of_view:
                raise ValueError(f"{scan_0} and {scan_1} have different field of view")
            # check distance
            if scan_0.distance is None:
                _logger.warning(f"no distance found for {scan_0}")
            elif not numpy.isclose(scan_0.distance, scan_1.distance, rtol=10e-3):
                raise ValueError(f"{scan_0} and {scan_1} have different sample / detector distance")
            # check pixel size
            if not numpy.isclose(scan_0.x_pixel_size, scan_1.x_pixel_size):
                raise ValueError(
                    f"{scan_0} and {scan_1} have different x pixel size. {scan_0.x_pixel_size} vs {scan_1.x_pixel_size}"
                )
            if not numpy.isclose(scan_0.y_pixel_size, scan_1.y_pixel_size):
                raise ValueError(
                    f"{scan_0} and {scan_1} have different y pixel size. {scan_0.y_pixel_size} vs {scan_1.y_pixel_size}"
                )
            # check magnification (only if is HDF5TomoScan)
            if isinstance(scan_0, HDF5TomoScan) and isinstance(scan_1, HDF5TomoScan):
                if not numpy.isclose(scan_0.magnification, scan_1.magnification):
                    raise ValueError(
                        f"{scan_0} and {scan_1} have different magnification. {scan_0.magnification} vs {scan_1.magnification}"
                    )
            if scan_0.dim_1 != scan_1.dim_1:
                raise ValueError(
                    f"projections width are expected to be the same. Not the canse for {scan_0} ({scan_0.dim_1} and {scan_1} ({scan_1.dim_1}))"
                )

        for scan in self.z_serie:
            # check x, y and z translation are constant (only if is an HDF5TomoScan)
            if isinstance(scan_0, HDF5TomoScan) and isinstance(scan_1, HDF5TomoScan):
                if scan.x_translation is not None and not numpy.isclose(
                    min(scan.x_translation), max(scan.x_translation)
                ):
                    _logger.warning(
                        "x translations appears to be evolving over time. Might end up with wrong stitching"
                    )
                if scan.y_translation is not None and not numpy.isclose(
                    min(scan.y_translation), max(scan.y_translation)
                ):
                    _logger.warning(
                        "y translations appears to be evolving over time. Might end up with wrong stitching"
                    )
                if scan.z_translation is not None and not numpy.isclose(
                    min(scan.z_translation), max(scan.z_translation)
                ):
                    _logger.warning(
                        "z translations appears to be evolving over time. Might end up with wrong stitching"
                    )

    def _compute_reduced_flats_and_darks(self):
        """
        TODO: should be done with nabu stuff !!!
        """
        for scan in self.z_serie:
            try:
                reduced_darks = scan.load_reduced_darks()
            except:
                _logger.info("no reduced dark found. Try to compute them.")
            if reduced_darks in (None, {}):
                reduced_darks = scan.compute_reduced_darks()
                try:
                    # if we don't have write in the folder containing the .nx for example
                    scan.save_reduced_darks(reduced_darks)
                except:
                    pass
            scan.set_reduced_darks(reduced_darks)

            try:
                reduced_flats = scan.load_reduced_flats()
            except:
                _logger.info("no reduced flats found. Try to compute them.")
            if reduced_flats in (None, {}):
                reduced_flats = scan.compute_reduced_flats()
                try:
                    # if we don't have write in the folder containing the .nx for example
                    scan.save_reduced_flats(reduced_flats)
                except:
                    pass
            scan.set_reduced_flats(reduced_flats)

    def _compute_shifts(self):
        """
        compute all shift requested (set to 'auto' in the configuration)
        """
        n_scans = len(self.configuration.input_scans)
        if n_scans == 0:
            raise ValueError("no scan to stich provided")

        # get shift
        final_shifts = []

        projection_for_shift = self.configuration.auto_relative_shift_params.get(KEY_CROSS_CORRELATION_SLICE, "middle")
        x_cross_correlation_function = self.configuration.auto_relative_shift_params.get(
            KEY_X_CROSS_CORRELATION_FUNC, None
        )
        y_cross_correlation_function = self.configuration.auto_relative_shift_params.get(
            KEY_Y_CROSS_CORRELATION_FUNC, None
        )

        for scan_0, scan_1, order_s0, order_s1, x_relative_shift, y_overlap in zip(
            self.z_serie[:-1],
            self.z_serie[1:],
            self.reading_order[:-1],
            self.reading_order[1:],
            self.configuration.x_shifts,
            self.configuration.overlap_height,
        ):
            # compute relative shift
            if x_relative_shift == "auto" or y_overlap == "auto":
                found_y, found_x = find_relative_shifts(
                    scan_0=scan_0,
                    scan_1=scan_1,
                    projection_for_shift=projection_for_shift,
                    x_cross_correlation_function=x_cross_correlation_function,
                    y_cross_correlation_function=y_cross_correlation_function,
                    invert_order=order_s1 != order_s0,
                    auto_flip=True,
                )

            final_shift = (
                found_y if y_overlap == "auto" else y_overlap,
                found_x if x_relative_shift == "auto" else x_relative_shift,
            )

            _logger.info(
                f"between {scan_0} and {scan_1} found a shift of {final_shift}. cross_correlation function used are: {x_cross_correlation_function} for x and {y_cross_correlation_function} for y"
            )
            final_shifts.append(final_shift)

        # set back values
        self.configuration.x_shifts = [final_shift[1] for final_shift in final_shifts]
        self.configuration.overlap_height = [abs(final_shift[0]) for final_shift in final_shifts]

    @staticmethod
    def _data_bunch_iterator(n_projections, bunch_size):
        proj_i = 0
        while n_projections - proj_i > bunch_size:
            yield (proj_i - bunch_size, proj_i)
            proj_i += bunch_size
        else:
            yield (proj_i, n_projections - 1)

    @staticmethod
    def _get_bunch_of_data(bunch_start: int, bunch_end: int, scans: tuple, scans_projections_indexes: tuple):
        """
        goal is to load contiguous projections as much as possible...

        :param scans_with_proj_indexes: tuple with scans and scan projection indexes to be loaded
        :return: list of list. For each frame we want to stitch contains the (flat fielded) frames to stich together
        """
        assert len(scans) == len(scans_projections_indexes)
        scans_proj_urls = []
        # for each scan store the real indices and the data url

        for scan, scan_projection_indexes in zip(scans, scans_projections_indexes):
            scan_proj_urls = {}
            # for each scan get the list of url to be loaded
            for i_proj in range(bunch_start, bunch_end):
                # for scan, scan_projections_indexes in zip(
                #     scans, scans_projections_indexes
                # ):
                proj_index_in_full_scan = scan_projection_indexes[i_proj]
                scan_proj_urls[proj_index_in_full_scan] = scan.projections[proj_index_in_full_scan]
            scans_proj_urls.append(scan_proj_urls)

        # then load data
        all_scan_final_data = numpy.empty((bunch_end - bunch_start, len(scans)), dtype=object)

        for i_scan, scan_urls in enumerate(scans_proj_urls):
            i_frame = 0
            _, set_of_compacted_slices = get_compacted_dataslices(scan_urls, return_url_set=True)
            for _, url in set_of_compacted_slices.items():
                url = DataUrl(
                    file_path=url.file_path(),
                    data_path=url.data_path(),
                    scheme="silx",
                    data_slice=url.data_slice(),
                )
                loaded_slices = get_data(url)
                if loaded_slices.ndim == 3:
                    n_slice = loaded_slices.shape[0]
                else:
                    n_slice = 1
                    loaded_slices = [
                        loaded_slices,
                    ]
                scan_indexes = list(scan_urls.keys())
                data = scan.flat_field_correction(
                    loaded_slices,
                    range(
                        scan_indexes[i_frame],
                        scan_indexes[i_frame] + n_slice,
                    ),
                )
                flip_lr = scans[i_scan].get_x_flipped(default=False)
                flip_ud = scans[i_scan].get_y_flipped(default=False)
                for frame in data:
                    f_frame = frame
                    if flip_lr:
                        f_frame = numpy.fliplr(f_frame)
                    if flip_ud:
                        f_frame = numpy.flipud(f_frame)

                    all_scan_final_data[i_frame, i_scan] = f_frame
                    i_frame += 1

        return all_scan_final_data

    def _create_nx_tomo(self):
        """
        create final NXtomo with stitched frames.
        Policy: save all projections flat fielded. So this NXtomo will only contain projections (no dark and no flat).
        But nabu will be able to reconstruct it with field `flatfield` set to False
        """
        if "auto" in self.configuration.x_shifts:
            raise RuntimeError("Looks like some shift haven't been computed")
        nx_tomo = NXtomo()

        nx_tomo.energy = self.z_serie[0].energy
        start_times = list(filter(None, [scan.start_time for scan in self.z_serie]))
        end_times = list(filter(None, [scan.end_time for scan in self.z_serie]))

        if len(start_times) > 0:
            nx_tomo.start_time = (
                numpy.asarray([numpy.datetime64(start_time) for start_time in start_times]).min().astype(datetime)
            )
        else:
            _logger.warning("Unable to find any start_time from input")
        if len(end_times) > 0:
            nx_tomo.end_time = (
                numpy.asarray([numpy.datetime64(end_time) for end_time in end_times]).max().astype(datetime)
            )
        else:
            _logger.warning("Unable to find any end_time from input")

        title = ",".join([scan.sequence_name or "" for scan in self.z_serie])
        nx_tomo.title = f"stitch done from {title}"

        # handle detector (without frames)
        nx_tomo.instrument.detector.field_of_view = self.z_serie[0].field_of_view
        nx_tomo.instrument.detector.distance = self.z_serie[0].distance
        nx_tomo.instrument.detector.x_pixel_size = self.z_serie[0].x_pixel_size
        nx_tomo.instrument.detector.y_pixel_size = self.z_serie[0].y_pixel_size
        nx_tomo.instrument.detector.image_key_control = [ImageKey.PROJECTION] * len(self.z_serie[0].projections)
        nx_tomo.instrument.detector.tomo_n = len(self.z_serie[0].projections)
        if isinstance(self.z_serie[0], HDF5TomoScan):
            nx_tomo.instrument.detector.magnification = self.z_serie[0].magnification
        # note: stitching process insure unflipping of frames
        nx_tomo.instrument.detector.x_flipped = False
        nx_tomo.instrument.detector.y_flipped = False

        if isinstance(self.z_serie[0], HDF5TomoScan):
            # note: first scan is always the reference as order to read data (so no rotation_angle inversion here)
            rotation_angle = numpy.asarray(self.z_serie[0].rotation_angle)
            nx_tomo.sample.rotation_angle = rotation_angle[
                numpy.asarray(self.z_serie[0].image_key_control) == ImageKey.PROJECTION.value
            ]
        elif isinstance(self.z_serie[0], EDFTomoScan):
            nx_tomo.sample.rotation_angle = numpy.linspace(
                start=0, stop=self.z_serie[0].scan_range, num=self.z_serie[0].tomo_n
            )
        else:
            raise NotImplementedError(
                f"scan type ({type(self.z_serie[0])} is not handled)",
                HDF5TomoScan,
                isinstance(self.z_serie[0], HDF5TomoScan),
            )

        # handle sample
        n_frames = len(nx_tomo.sample.rotation_angle)
        if False not in [isinstance(scan, HDF5TomoScan) for scan in self.z_serie]:
            # we consider the new x, y and z position to be at the center of the one created
            x_translation = [scan.x_translation for scan in self.z_serie if scan.x_translation is not None]
            nx_tomo.sample.x_translation = [numpy.asarray(x_translation).mean()] * n_frames
            y_translation = [scan.y_translation for scan in self.z_serie if scan.y_translation is not None]
            nx_tomo.sample.y_translation = [numpy.asarray(y_translation).mean()] * n_frames
            z_translation = [scan.z_translation for scan in self.z_serie if scan.z_translation is not None]
            nx_tomo.sample.z_translation = [numpy.asarray(z_translation).mean()] * n_frames

            nx_tomo.sample.name = self.z_serie[0].sample_name

        # compute stiched frame shape
        n_proj = len(self.z_serie[0].projections)
        y_overlaps = self.configuration.overlap_height
        stitched_frame_shape = (
            n_proj,
            int(
                numpy.asarray([scan.dim_2 for scan in self.z_serie]).sum()
                - numpy.asarray([abs(overlap) for overlap in y_overlaps]).sum()
            ),
            self.z_serie[0].dim_1,
        )

        # get expected output dataset first (just in case output and input files are the same)
        first_proj_idx = sorted(self.z_serie[0].projections.keys())[0]
        first_proj_url = self.z_serie[0].projections[first_proj_idx]
        if h5py.is_hdf5(first_proj_url.file_path()):
            first_proj_url = DataUrl(
                file_path=first_proj_url.file_path(),
                data_path=first_proj_url.data_path(),
                scheme="h5py",
            )

        # first save the NXtomo entry without the frame
        # dicttonx will fail if the folder does not exists
        os.makedirs(os.path.dirname(self.configuration.output_file_path), exist_ok=True)
        nx_tomo.save(
            file_path=self.configuration.output_file_path,
            data_path=self.configuration.output_data_path,
            nexus_path_version=self.configuration.output_nexus_version,
            overwrite=self.configuration.overwrite_results,
        )

        _logger.info(
            f"reading order is {self.reading_order}",
        )

        # append frames ("instrument/detactor/data" dataset)
        with HDF5File(filename=self.configuration.output_file_path, mode="a") as h5f:
            # note: nx_tomo.save already handles the possible overwrite conflict by removing
            # self.configuration.output_file_path or raising an error

            stitched_frame_path = "/".join(
                [
                    self.configuration.output_data_path,
                    _get_nexus_paths(self.configuration.output_nexus_version).PROJ_PATH,
                ]
            )

            projection_dataset = h5f.create_dataset(
                name=stitched_frame_path,
                shape=stitched_frame_shape,
                dtype=self.configuration.output_dtype,
            )
            # TODO: we could also create in several time and create a virtual dataset from it.
            scans_projections_indexes = []
            for scan, reverse in zip(self.z_serie, self.reading_order):
                scans_projections_indexes.append(sorted(scan.projections.keys(), reverse=(reverse == -1)))
            if self.progress:
                self.progress.set_max_advancement(len(scan.projections.keys()))

            # for each indexes create a value which is the list of url to stitch together
            # for now only try to do the first two
            overlap_kernels = []
            for overlap_height in self.configuration.overlap_height:
                overlap_kernels.append(
                    ZStichOverlapKernel(
                        frame_width=self.z_serie[0].dim_1,
                        stitching_strategy=self.configuration.stitching_strategy,
                        overlap_height=-1 if overlap_height == "auto" else overlap_height,
                    )
                )

            i_proj = 0
            for bunch_start, bunch_end in self._data_bunch_iterator(len(scan.projections), bunch_size=50):
                for data_frames in self._get_bunch_of_data(
                    bunch_start,
                    bunch_end,
                    scans=self.z_serie,
                    scans_projections_indexes=scans_projections_indexes,
                ):
                    # TODO: try to do this in parallel or at least dump then in one go. but not sure this last one would speed up.
                    # should be handled by the flushing mecanism
                    ZStitcher.stitch_frames(
                        frames=data_frames,
                        x_relative_shifts=self.configuration.x_shifts,
                        y_overlap_heights=self.configuration.overlap_height,
                        output_dataset=projection_dataset,
                        overlap_kernels=overlap_kernels,
                        i_frame=i_proj,
                        output_dtype=self.configuration.output_dtype,
                    )
                    if self.progress is not None:
                        self.progress.increase_advancement()
                    i_proj += 1

            # create link to this dataset that can be missing
            # "data/data" link
            if "data" in h5f[self.configuration.output_data_path]:
                data_group = h5f[self.configuration.output_data_path]["data"]
                if not stitched_frame_path.startswith("/"):
                    stitched_frame_path = "/" + stitched_frame_path
                data_group["data"] = h5py.SoftLink(stitched_frame_path)
                if "default" not in h5f[self.configuration.output_data_path].attrs:
                    h5f[self.configuration.output_data_path].attrs["default"] = "data"
                for attr_name, attr_value in zip(
                    ("NX_class", "SILX_style/axis_scale_types", "signal"),
                    ("NXdata", ["linear", "linear"], "data"),
                ):
                    if attr_name not in data_group.attrs:
                        data_group.attrs[attr_name] = attr_value

        return nx_tomo

    def _dump_stitching_configuration(self):
        """dump configuration used for stitching at the NXtomo entry"""
        process_name = "stitching_configuration"
        config_dict = self.configuration.to_dict()
        # adding nabu specific information
        nabu_process_info = {
            "@NX_class": "NXentry",
            f"{process_name}@NX_class": "NXprocess",
            f"{process_name}/program": "nabu-stitching",
            f"{process_name}/version": nabu_version,
            f"{process_name}/date": get_datetime(),
            f"{process_name}/configuration": config_dict,
        }

        dicttonx(
            nabu_process_info,
            h5file=self.configuration.output_file_path,
            h5path=self.configuration.output_data_path,
            update_mode="replace",
            mode="a",
        )


def z_stitch_raw_frames(
    frames: tuple,
    y_shifts: tuple,
    output_dtype: numpy.dtype = numpy.float32,
    check_inputs=True,
    overlap_kernels: Optional[Union[ZStichOverlapKernel, tuple]] = None,
    raw_frames_compositions: Optional[ZFrameComposition] = None,
    overlap_frames_compositions: Optional[ZFrameComposition] = None,
) -> numpy.ndarray:
    """
    stitches raw frames (already shifted and flat fielded !!!) together using
    raw stitching (no pixel interpolation, y_overlap_in_px is expected to be a int)
    returns stitched_projection, raw_img_1, raw_img_2, computed_overlap
    proj_0 and pro_1 are already expected to be in a row. Having stitching_height_in_px in common. At top of proj_0
    and at bottom of proj_1

    :param tuple frames: tuple of 2D numpy array
    :param stitching_heights_in_px: scalar value of the stitching to apply or a list of len(frames) - 1 size with each stitching_height to apply between each couple of frame
    :param raw_frames_compositions: pre computed raw frame composition. If not provided will compute them. allow providing it to speed up calculation
    :param overlap_frames_compositions: pre computed stitched frame composition. If not provided will compute them. allow providing it to speed up calculation
    :param overlap_kernels: ZStichOverlapKernel overlap kernel to be used or a list of kernel (one per overlap). Define startegy and overlap heights
    :param numpy.dtype output_dtype: dataset dtype. For now must be provided because flat field corrcetion change data type (numpy.float32 for now)
    """
    if overlap_kernels is None:
        # handle overlap area
        if overlap_kernels is None and len(frames) > 0:
            # FIXME !
            # pylint: disable= E1123,E1120
            proj_0 = frames[0]
            overlap_kernels = ZStichOverlapKernel(
                stitching_strategy=DEFAULT_OVERLAP_STRATEGY,
                overlap_height=proj_0.shape[0],
                frame_width=proj_0.shape[1],
            )
        if isinstance(overlap_kernels, OverlapKernelBase):
            overlap_kernels = [copy(overlap_kernels) for _ in (len(frames) - 1)]

    if check_inputs:

        def check_proj(proj):
            if not isinstance(proj, numpy.ndarray) and proj.ndim == 2:
                raise ValueError(f"frames are expected to be 2D numpy array")

        [check_proj(frame) for frame in frames]

        for proj_0, proj_1 in zip(frames[:-1], frames[1:]):
            if proj_0.shape[1] != proj_1.shape[1]:
                raise ValueError("Both projections are expected to have the same width")
        for proj_0, proj_1, kernel in zip(frames[:-1], frames[1:], overlap_kernels):
            if proj_0.shape[0] <= kernel.overlap_height:
                raise ValueError(
                    f"proj_0 height ({proj_0.shape[0]}) is less than kernel overlap ({kernel.overlap_height})"
                )
            if proj_1.shape[0] <= kernel.overlap_height:
                raise ValueError(
                    f"proj_1 height ({proj_1.shape[0]}) is less than kernel overlap ({kernel.overlap_height})"
                )

    # cast shift in int: for now only case handled
    y_shifts = [int(y_shift) for y_shift in y_shifts]
    # step 0: create numpy array that will contain stitching
    stitched_projection_shape = (
        # here we only handle frames because shift are already done
        # + because shift are expected to be negative
        int(
            numpy.asarray([frame.shape[0] for frame in frames]).sum() + numpy.asarray(y_shifts).sum(),
        ),
        frames[0].shape[1],
    )
    stitch_array = numpy.empty(stitched_projection_shape, dtype=output_dtype)

    # step 1: set kernel overlap height if undefined
    for y_shift, kernel in zip(y_shifts, overlap_kernels):
        if kernel.overlap_height in (-1, None):
            kernel.overlap_height = abs(y_shift)

    # step 2: set raw data

    # fill stitch array with raw data raw data
    if raw_frames_compositions is None:
        raw_frames_compositions = ZFrameComposition.compute_raw_frame_compositions(
            frames=frames,
            overlap_kernels=overlap_kernels,
            y_shifts=y_shifts,
        )
    raw_frames_compositions.compose(
        output_frame=stitch_array,
        input_frames=frames,
    )

    # step 3 set stitched data

    # 3.1 create stitched overlaps
    stitched_overlap = []
    for frame_0, frame_1, kernel, y_shift in zip(frames[:-1], frames[1:], overlap_kernels, y_shifts):
        if y_shift >= 0:
            raise ValueError("No overlap found. Unagle to do stitching on it")

        frame_0_overlap, frame_1_overlap = ZStitcher.get_overlap_areas(
            frame_0,
            frame_1,
            real_overlap=abs(y_shift),
            stitching_height=kernel.overlap_height,
        )
        assert (
            frame_0_overlap.shape[0] == frame_1_overlap.shape[0] == kernel.overlap_height
        ), f"{frame_0_overlap.shape[0]} == {frame_1_overlap.shape[0]} == {kernel.overlap_height}"

        stitched_overlap.append(
            kernel.stitch(
                frame_0_overlap,
                frame_1_overlap,
            )[0]
        )
    # 3.2 fill stitched overlap on output array
    if overlap_frames_compositions is None:
        overlap_frames_compositions = ZFrameComposition.compute_stitch_frame_composition(
            frames=frames,
            overlap_kernels=overlap_kernels,
            y_shifts=y_shifts,
        )
    overlap_frames_compositions.compose(
        output_frame=stitch_array,
        input_frames=stitched_overlap,
    )

    return stitch_array



from numpy import prod

try:
    import pyopencl as cl

    __has_pyopencl__ = True
    __pyopencl_error_msg__ = None
except ImportError as err:
    __has_pyopencl__ = False
    __pyopencl_error_msg__ = str(err)
from ..resources.gpu import GPUDescription


def usable_opencl_devices():
    """
    Test the available OpenCL platforms/devices.

    Returns
    --------
    platforms: dict
        Dictionary where the key is the platform name, and the value is a list
        of `silx.opencl.common.Device` object.
    """
    platforms = {}
    for platform in cl.get_platforms():
        platforms[platform.name] = platform.get_devices()
    return platforms


def detect_opencl_gpus():
    """
    Get the available OpenCL-compatible GPUs.

    Returns
    --------
    gpus: dict
        Nested dictionary where the keys are OpenCL platform names,
        values are dictionary of GPU IDs and `silx.opencl.common.Device` object.
    error_msg: str
        In the case where there is an error, the message is returned in this item.
        Otherwise, it is a None object.
    """
    gpus = {}
    error_msg = None
    if not (__has_pyopencl__):
        return {}, __pyopencl_error_msg__
    try:
        platforms = usable_opencl_devices()
    except Exception as exc:
        error_msg = str(exc)
    if error_msg is not None:
        return {}, error_msg
    for platform_name, devices in platforms.items():
        for d_id, device in enumerate(devices):
            if device.type == cl.device_type.GPU:  # and bool(device.available):
                if platform_name not in gpus:
                    gpus[platform_name] = {}
                gpus[platform_name][d_id] = device
    return gpus, None


def collect_opencl_gpus():
    """
    Return a dictionary of platforms and brief description of each OpenCL-compatible
    GPU with a few fields
    """
    gpus, error_msg = detect_opencl_gpus()
    if error_msg is not None:
        return None
    opencl_gpus = {}
    for platform, gpus in gpus.items():
        for gpu_id, gpu in gpus.items():
            if platform not in opencl_gpus:
                opencl_gpus[platform] = {}
            opencl_gpus[platform][gpu_id] = GPUDescription(gpu, device_id=gpu_id).get_dict()
            opencl_gpus[platform][gpu_id]["platform"] = platform
    return opencl_gpus


def collect_opencl_cpus():
    """
    Return a dictionary of platforms and brief description of each OpenCL-compatible
    CPU with a few fields
    """
    opencl_cpus = {}
    platforms = usable_opencl_devices()
    for platform, devices in platforms.items():
        if "cuda" in platform.lower():
            continue
        opencl_cpus[platform] = {}
        for device_id, device in enumerate(devices):  # device_id might be inaccurate
            if device.type != cl.device_type.CPU:
                continue
            opencl_cpus[platform][device_id] = GPUDescription(device).get_dict()
            opencl_cpus[platform][device_id]["platform"] = platform
    return opencl_cpus


def create_opencl_context(platform_id, device_id, cleanup_at_exit=True):
    """
    Create an OpenCL context.
    """
    platforms = cl.get_platforms()
    platform = platforms[platform_id]
    devices = platform.get_devices()
    ctx = cl.Context(devices=[devices[device_id]])
    return ctx


def replace_array_memory(arr, new_shape):
    """
    Replace the underlying buffer data of a  `pyopencl.array.Array`.
    This function is dangerous !
    It should merely be used to clear memory, the array should not be used afterwise.
    """
    arr.data.release()
    arr.base_data = cl.Buffer(arr.context, cl.mem_flags.READ_WRITE, prod(new_shape) * arr.dtype.itemsize)
    arr.shape = new_shape
    # strides seems to be updated by pyopencl
    return arr


def pick_opencl_cpu_platform(opencl_cpus):
    """
    Pick the best OpenCL implementation for the opencl cpu.
    This function assume that there is only one opencl-enabled CPU on the
    current machine, but there might be several OpenCL implementations/vendors.


    Parameters
    ----------
    opencl_cpus: dict
        Dictionary with the available opencl-enabled CPUs.
        Usually obtained with collect_opencl_cpus().

    Returns
    -------
    cpu: dict
        A dictionary describing the CPU.
    """
    if len(opencl_cpus) == 0:
        raise ValueError("No CPU to pick")
    name2device = {}
    for platform, devices in opencl_cpus.items():
        for device_id, device_desc in devices.items():
            name2device.setdefault(device_desc["name"], [])
            name2device[device_desc["name"]].append(platform)
    if len(name2device) > 1:
        raise ValueError("Expected at most one CPU but got %d: %s" % (len(name2device), list(name2device.keys())))
    cpu_name = list(name2device.keys())[0]
    platforms = name2device[cpu_name]
    # Several platforms for the same CPU
    res = opencl_cpus[platforms[0]]
    if len(platforms) > 1:
        if "intel" in cpu_name.lower():
            for platform in platforms:
                if "intel" in platform.lower():
                    res = opencl_cpus[platform]
    #
    return res[list(res.keys())[0]]



import numpy as np
from scipy import sparse


class DetectorDistortionBase:
    """ """

    def __init__(self, detector_full_shape_vh=(0, 0)):
        """This is the basis class.
        A simple identity transformation which has the only merit to show
        how it works.Reimplement this function to have more parameters for other
        transformations
        """
        self._build_full_transformation(detector_full_shape_vh)

    def transform(self, source_data, do_full=False):
        """performs the transformation"""

        if do_full:
            result = self.transformation_matrix_full @ source_data.flat
            result.shape = source_data.shape
        else:
            result = self.transformation_matrix @ source_data.flat
            result.shape = self.target_shape

        return result

    def _build_full_transformation(self, detector_full_shape_vh):
        """A simple identity.
        Reimplement this function to have more parameters for other
        transformations
        """

        self.detector_full_shape_vh = detector_full_shape_vh

        sz, sx = detector_full_shape_vh

        # A simple identity matrix in sparse coordinates format

        self.total_detector_npixs = detector_full_shape_vh[0] * detector_full_shape_vh[1]
        I_tmp = np.arange(self.total_detector_npixs)
        J_tmp = np.arange(self.total_detector_npixs)
        V_tmp = np.ones([self.total_detector_npixs], "f")

        coo_tmp = sparse.coo_matrix((V_tmp, (I_tmp, J_tmp)), shape=(sz * sx, sz * sx))

        csr_tmp = coo_tmp.tocsr()

        ## The following arrays  are kept for future usage
        ## when, according to the "sub_region" parameter of the moment,
        ## they will be used to extract a "slice" of them
        ## which will map an appropriate data region corresponding to "sub_region_source"
        ## to the target "sub_region" of the moment
        self.full_csr_data = csr_tmp.data
        self.full_csr_indices = csr_tmp.indices
        self.full_csr_indptr = csr_tmp.indptr

        ## This will be used to save time if the same sub_region argument is requested several time in a row
        self._status = None

    def get_adapted_subregion(self, sub_region_xz):
        if sub_region_xz is not None:
            start_x, end_x, start_z, end_z = sub_region_xz
        else:
            start_x = 0
            end_x = None
            start_z = 0
            end_z = None

        (start_x, end_x, start_z, end_z) = self.set_sub_region_transformation((start_x, end_x, start_z, end_z))
        return (start_x, end_x, start_z, end_z)

    def set_sub_region_transformation(self, target_sub_region=None):
        """must return a source sub_region. It sets internally  an object (a practical implementation
        would be a sparse matrice) which can be reused in further applications of "transform" method
        for transforming the source sub_region data into the target sub_region
        """
        if target_sub_region is None:
            target_sub_region = (None, None, 0, None)

        if self._status is not None and self._status["target_sub_region"] == target_sub_region:
            return self._status["source_sub_region"]
        else:
            self._status = None
            return self._set_sub_region_transformation(target_sub_region)

    def set_full_transformation(self):
        self._set_sub_region_transformation(do_full=True)

    def get_actual_shapes_source_target(self):
        if self._status is None:
            return None, None
        else:
            return self._status["source_sub_region"], self._status["target_sub_region"]

    def _set_sub_region_transformation(self, target_sub_region=None, do_full=False):
        """to be reimplemented in the derived classes"""
        if target_sub_region is None or do_full:
            target_sub_region = (None, None, 0, None)

        (x_start, x_end, z_start, z_end) = target_sub_region

        if z_start is None:
            z_start = 0
        if z_end is None:
            z_end = self.detector_full_shape_vh[0]

        if (x_start, x_end) not in [(None, None), (0, None), (0, self.detector_full_shape_vh[1])]:
            message = f""" In the base class DetectorDistortionBase only vertical slicing is accepted.
            The sub_region contained (x_start, x_end)={(x_start, x_end)} which would slice the 
            full horizontal size which is {self.detector_full_shape_vh[1]}
            """
            raise ValueError()

        x_start, x_end = 0, self.detector_full_shape_vh[1]

        row_ptr_start = z_start * self.detector_full_shape_vh[1]
        row_ptr_end = z_end * self.detector_full_shape_vh[1]

        indices_start = self.full_csr_indptr[row_ptr_start]
        indices_end = self.full_csr_indptr[row_ptr_end]

        indices_offset = self.full_csr_indptr[row_ptr_start]

        source_offset = target_sub_region[2] * self.detector_full_shape_vh[1]

        data_tmp = self.full_csr_data[indices_start:indices_end]
        indices_tmp = self.full_csr_indices[indices_start:indices_end] - source_offset
        indptr_tmp = self.full_csr_indptr[row_ptr_start : row_ptr_end + 1] - indices_offset

        target_size = (z_end - z_start) * self.detector_full_shape_vh[1]
        source_size = (z_end - z_start) * self.detector_full_shape_vh[1]

        tmp_transformation_matrix = sparse.csr_matrix(
            (data_tmp, indices_tmp, indptr_tmp), shape=(target_size, source_size)
        )

        if do_full:
            self.transformation_matrix_full = tmp_transformation_matrix
            return None
        else:
            self.transformation_matrix = tmp_transformation_matrix

        self.target_shape = ((z_end - z_start), self.detector_full_shape_vh[1])

        ## For the identity matrix the source and the target have the same size.
        ## The two following lines are trivial.
        ## For this identity transformation only the slicing of the appropriate part
        ## of the identity sparse matrix is slightly laborious.
        ## Practical case will be more complicated and source_sub_region
        ## will be in general larger than the target_sub_region
        self._status = {
            "target_sub_region": ((x_start, x_end, z_start, z_end)),
            "source_sub_region": ((x_start, x_end, z_start, z_end)),
        }

        return self._status["source_sub_region"]


class DetectorDistortionMapsXZ(DetectorDistortionBase):
    def __init__(self, map_x, map_z):
        """
        This class implements the distortion correction from the knowledge of
        two arrays, map_x and map_z.
        Pixel (i,j) of the corrected image is obtained by interpolating the raw data at position
        ( map_z(i,j), map_x(i,j) ).

        Parameters:
            map_x : float 2D array
            map_z : float 2D array
        """

        self._build_full_transformation(map_x, map_z)

    def _build_full_transformation(self, map_x, map_z):
        """ """
        detector_full_shape_vh = map_x.shape
        if detector_full_shape_vh != map_z.shape:
            message = f"""  map_x and map_z must have the same shape
            but the dimensions were {map_x.shape} and {map_z.shape}
            """
            raise ValueError(message)

        coordinates = np.array([map_z, map_x])

        # padding
        sz, sx = detector_full_shape_vh
        total_detector_npixs = sz * sx
        xs = np.clip(np.array(coordinates[1].flat), [[0]], [[sx - 1]])
        zs = np.clip(np.array(coordinates[0].flat), [[0]], [[sz - 1]])

        ix0s = np.floor(xs)
        ix1s = np.ceil(xs)
        fx = xs - ix0s

        iz0s = np.floor(zs)
        iz1s = np.ceil(zs)
        fz = zs - iz0s

        I_tmp = np.empty([4 * sz * sx], np.int64)
        J_tmp = np.empty([4 * sz * sx], np.int64)
        V_tmp = np.ones([4 * sz * sx], "f")

        I_tmp[:] = np.arange(sz * sx * 4) // 4

        J_tmp[0::4] = iz0s * sx + ix0s
        J_tmp[1::4] = iz0s * sx + ix1s
        J_tmp[2::4] = iz1s * sx + ix0s
        J_tmp[3::4] = iz1s * sx + ix1s

        V_tmp[0::4] = (1 - fz) * (1 - fx)
        V_tmp[1::4] = (1 - fz) * fx
        V_tmp[2::4] = fz * (1 - fx)
        V_tmp[3::4] = fz * fx

        self.detector_full_shape_vh = detector_full_shape_vh

        coo_tmp = sparse.coo_matrix((V_tmp.astype("f"), (I_tmp, J_tmp)), shape=(sz * sx, sz * sx))

        csr_tmp = coo_tmp.tocsr()

        self.full_csr_data = csr_tmp.data
        self.full_csr_indices = csr_tmp.indices
        self.full_csr_indptr = csr_tmp.indptr

        ## This will be used to save time if the same sub_region argument is requested several time in a row
        self._status = None

    def _set_sub_region_transformation(
        self,
        target_sub_region=(
            (
                None,
                None,
                0,
                0,
            ),
        ),
        do_full=False,
    ):
        if target_sub_region is None or do_full:
            target_sub_region = (None, None, 0, None)

        (x_start, x_end, z_start, z_end) = target_sub_region

        if z_start is None:
            z_start = 0
        if z_end is None:
            z_end = self.detector_full_shape_vh[0]

        if (x_start, x_end) not in [(None, None), (0, None), (0, self.detector_full_shape_vh[1])]:
            message = f""" In the base class DetectorDistortionRotation only vertical slicing is accepted.
            The sub_region contained (x_start, x_end)={(x_start, x_end)} which would slice the 
            full horizontal size which is {self.detector_full_shape_vh[1]}
            """
            raise ValueError()

        x_start, x_end = 0, self.detector_full_shape_vh[1]

        row_ptr_start = z_start * self.detector_full_shape_vh[1]
        row_ptr_end = z_end * self.detector_full_shape_vh[1]

        indices_start = self.full_csr_indptr[row_ptr_start]
        indices_end = self.full_csr_indptr[row_ptr_end]

        data_tmp = self.full_csr_data[indices_start:indices_end]

        target_offset = self.full_csr_indptr[row_ptr_start]
        indptr_tmp = self.full_csr_indptr[row_ptr_start : row_ptr_end + 1] - target_offset

        indices_tmp = self.full_csr_indices[indices_start:indices_end]

        iz_source = (indices_tmp) // self.detector_full_shape_vh[1]

        z_start_source = iz_source.min()
        z_end_source = iz_source.max() + 1
        source_offset = z_start_source * self.detector_full_shape_vh[1]
        indices_tmp = indices_tmp - source_offset

        target_size = (z_end - z_start) * self.detector_full_shape_vh[1]
        source_size = (z_end_source - z_start_source) * self.detector_full_shape_vh[1]

        tmp_transformation_matrix = sparse.csr_matrix(
            (data_tmp, indices_tmp, indptr_tmp), shape=(target_size, source_size)
        )

        if do_full:
            self.transformation_matrix_full = tmp_transformation_matrix
            return None
        else:
            self.transformation_matrix = tmp_transformation_matrix

        self.target_shape = ((z_end - z_start), self.detector_full_shape_vh[1])

        ## For the identity matrix the source and the target have the same size.
        ## The two following lines are trivial.
        ## For this identity transformation only the slicing of the appropriate part
        ## of the identity sparse matrix is slightly laborious.
        ## Practical case will be more complicated and source_sub_region
        ## will be in general larger than the target_sub_region
        self._status = {
            "target_sub_region": ((x_start, x_end, z_start, z_end)),
            "source_sub_region": ((x_start, x_end, z_start_source, z_end_source)),
        }
        return self._status["source_sub_region"]



from copy import copy
from dataclasses import dataclass

import numpy

from nabu.stitching.overlap import ZStichOverlapKernel


@dataclass
class _FrameCompositionBase:
    def compose(self, output_frame: numpy.ndarray, input_frames: tuple):
        raise NotImplementedError("Base class")


@dataclass
class ZFrameComposition(_FrameCompositionBase):
    """
    class used to define intervals to know where to dump raw data or stitched data according to requested policy.
    The idea is to create this once for all for one stitching operation and reuse it for each frame.
    """

    local_start_y: tuple
    local_end_y: tuple
    global_start_y: tuple
    global_end_y: tuple

    def browse(self):
        for i in range(len(self.local_start_y)):
            yield (
                self.local_start_y[i],
                self.local_end_y[i],
                self.global_start_y[i],
                self.global_end_y[i],
            )

    def compose(self, output_frame: numpy.ndarray, input_frames: tuple):
        if not output_frame.ndim == 2:
            raise TypeError(f"output_frame is expected to be 2D and not {output_frame.ndim}")
        for (
            global_start_y,
            global_end_y,
            local_start_y,
            local_end_y,
            input_frame,
        ) in zip(
            self.global_start_y,
            self.global_end_y,
            self.local_start_y,
            self.local_end_y,
            input_frames,
        ):
            if input_frame is not None:
                output_frame[global_start_y:global_end_y] = input_frame[local_start_y:local_end_y]

    @staticmethod
    def compute_raw_frame_compositions(frames: tuple, y_shifts: tuple, overlap_kernels: tuple):
        """
        compute frame composition for raw data
        """
        assert len(frames) == len(overlap_kernels) + 1 == len(y_shifts) + 1

        global_start_y = []
        local_start_y = []
        global_end_y = []
        local_end_y = []
        frame_height_sum = 0

        # extend shifts and kernels to have a first shift of 0 and two overlaps values at 0 to
        # generalize processing
        lower_shifts = [0]
        lower_shifts.extend(y_shifts)
        upper_shifts = list(copy(y_shifts))
        upper_shifts.append(0)
        overlaps = [kernel.overlap_height for kernel in overlap_kernels]
        overlaps.append(0)
        overlaps.insert(0, 0)

        for (
            frame,
            lower_shift,
            upper_shift,
            lower_overlap_kernel,
            upper_overlap_kernel,
        ) in zip(frames, lower_shifts, upper_shifts, overlaps[:-1], overlaps[1:]):
            if lower_shift > 0:
                raise ValueError(
                    "Unexpected shift value found; TODO: handle positive shift (this mean no overlap between frames"
                )
            lower_remaining = abs(lower_shift) - lower_overlap_kernel
            upper_o_s_diff = abs(upper_shift) - upper_overlap_kernel
            # policy with overlap that needs to take one more line on one side ((lower_remaining) % 2 == 1)
            # take this line on the lower frame side.
            new_local_start_y = lower_overlap_kernel + (lower_remaining) // 2 - (lower_remaining) % 2
            new_local_end_y = frame.shape[0] - (upper_overlap_kernel + (upper_o_s_diff) // 2)

            new_global_start_y = frame_height_sum + new_local_start_y
            new_global_end_y = frame_height_sum + new_local_end_y
            frame_height_sum += frame.shape[0] - abs(upper_shift)

            # check values are coherent
            if new_local_start_y < 0 or new_local_end_y < 0 or new_global_start_y < 0 or new_global_end_y < 0:
                raise ValueError(
                    "Incoherence found on the computing raw frame composition. Are you sure overlap height is no larger than frame height"
                )
            global_start_y.append(int(new_global_start_y))
            global_end_y.append(int(new_global_end_y))
            local_start_y.append(int(new_local_start_y))
            local_end_y.append(int(new_local_end_y))

        return ZFrameComposition(
            local_start_y=tuple(local_start_y),
            local_end_y=tuple(local_end_y),
            global_start_y=tuple(global_start_y),
            global_end_y=tuple(global_end_y),
        )

    @staticmethod
    def compute_stitch_frame_composition(frames, y_shifts: tuple, overlap_kernels: tuple):
        """
        compute frame composition for stiching.
        """
        assert len(frames) == len(overlap_kernels) + 1
        global_start_y = []
        global_end_y = []
        local_start_y = [0] * len(y_shifts)  # stiched is expected to be at the expected size already
        local_end_y = [kernel.overlap_height for kernel in overlap_kernels]

        frame_height_sum = 0
        for frame, kernel, y_shift in zip(frames[:-1], overlap_kernels, y_shifts):
            if y_shift > 0:
                raise ValueError(
                    "Unexpected shift value found; TODO: handle positive shift (this mean no overlap between frames"
                )

            assert isinstance(kernel, ZStichOverlapKernel)
            new_global_start_y = (
                frame_height_sum
                + frame.shape[0]
                - (kernel.overlap_height + (abs(y_shift) - kernel.overlap_height) // 2)
            )
            new_global_end_y = frame_height_sum + frame.shape[0] - (abs(y_shift) - kernel.overlap_height) // 2
            if new_global_start_y < 0 or new_global_end_y < 0:
                raise ValueError(
                    "Incoherence found on the computing raw frame composition. Are you sure overlap height is no larger than frame height"
                )
            global_start_y.append(int(new_global_start_y))
            global_end_y.append(int(new_global_end_y))
            frame_height_sum += frame.shape[0] + y_shift

        return ZFrameComposition(
            local_start_y=tuple(local_start_y),
            local_end_y=tuple(local_end_y),
            global_start_y=tuple(global_start_y),
            global_end_y=tuple(global_end_y),
        )

    @staticmethod
    def pprint_z_stitching(raw_composition, stitch_composition):
        """
        util to display what the output of the z stitch will looks like from composition
        """
        for i_frame, (raw_comp, stitch_comp) in enumerate(zip(raw_composition.browse(), stitch_composition.browse())):
            raw_local_start, raw_local_end, raw_global_start, raw_global_end = raw_comp

            print(
                f"stitch_frame[{raw_global_start}:{raw_global_end}] = frame_{i_frame}[{raw_local_start}:{raw_local_end}]"
            )

            (
                stitch_local_start,
                stitch_local_end,
                stitch_global_start,
                stitch_global_end,
            ) = stitch_comp

            print(
                f"stitch_frame[{stitch_global_start}:{stitch_global_end}] = stitched_frame_{i_frame}[{stitch_local_start}:{stitch_local_end}]"
            )
        else:
            i_frame += 1
            raw_local_start, raw_local_end, raw_global_start, raw_global_end = list(raw_composition.browse())[-1]
            print(
                f"stitch_frame[{raw_global_start}:{raw_global_end}] = frame_{i_frame}[{raw_local_start}:{raw_local_end}]"
            )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import argparse
import sys
import os
import h5py
import tomoscan.validator
from tomoscan.esrf.scan.hdf5scan import HDF5TomoScan
from tomoscan.esrf.scan.edfscan import EDFTomoScan


def get_scans(path, entries: str):
    path = os.path.abspath(path)
    res = []
    if EDFTomoScan.is_tomoscan_dir(path):
        res.append(EDFTomoScan(scan=path))
    elif HDF5TomoScan.is_tomoscan_dir(path):
        if entries == "__all__":
            entries = HDF5TomoScan.get_valid_entries(path)
        for entry in entries:
            res.append(HDF5TomoScan(path, entry))
    else:
        raise TypeError(f"{path} does not looks like a folder containing .EDF or a valid nexus file ")
    return res


def main():
    argv = sys.argv
    parser = argparse.ArgumentParser(description="Check if provided scan(s) seems valid to be reconstructed.")
    parser.add_argument("path", help="Data to validate (h5 file, edf folder)")
    parser.add_argument("entries", help="Entries to be validated (in the case of a h5 file)", nargs="*")
    parser.add_argument(
        "--ignore-dark",
        help="Do not check for dark",
        default=True,
        action="store_false",
        dest="check_dark",
    )
    parser.add_argument(
        "--ignore-flat",
        help="Do not check for flat",
        default=True,
        action="store_false",
        dest="check_flat",
    )
    parser.add_argument(
        "--no-phase-retrieval",
        help="Check scan energy, distance and pixel size",
        dest="check_phase_retrieval",
        default=True,
        action="store_false",
    )
    parser.add_argument(
        "--check-nan",
        help="Check frames if contains any nan.",
        dest="check_nan",
        default=False,
        action="store_true",
    )
    parser.add_argument(
        "--skip-links-check",
        "--no-link-check",
        help="Check frames dataset if have some broken links.",
        dest="check_vds",
        default=True,
        action="store_false",
    )
    parser.add_argument(
        "--all-entries",
        help="Check all entries of the files (for HDF5 only for now)",
        default=False,
        action="store_true",
    )
    parser.add_argument(
        "--extend",
        help="By default it only display items with issues. Extend will display them all",
        dest="only_issues",
        default=True,
        action="store_false",
    )

    options = parser.parse_args(argv[1:])
    if options.all_entries is True:
        entries = "__all__"
    else:
        if len(options.entries) == 0 and h5py.is_hdf5(options.path):
            entries = "__all__"
        else:
            entries = options.entries

    scans = get_scans(path=options.path, entries=entries)
    if len(scans) == 0:
        raise ValueError(f"No scan found from file:{options.path}, entries:{options.entries}")
    for scan in scans:
        validator = tomoscan.validator.ReconstructionValidator(
            scan=scan,
            check_phase_retrieval=options.check_phase_retrieval,
            check_values=options.check_nan,
            check_vds=options.check_vds,
            check_dark=options.check_dark,
            check_flat=options.check_flat,
        )
        sys.stdout.write(validator.checkup(only_issues=options.only_issues))


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import os
import pytest
from nabu.utils import get_folder_path
from nabu import __nabu_modules__ as nabu_modules


def get_modules_to_test(mods):
    sep = os.sep
    modules = []
    extra_args = []
    for mod in mods:
        if mod.startswith("-"):
            extra_args.append(mod)
            continue
        # Test a whole module
        if mod.lower() in nabu_modules:
            mod_abspath = os.path.join(get_folder_path(mod), "tests")
        # Test an individual file
        else:
            mod_path = mod.replace(".", sep) + ".py"
            mod_abspath = get_folder_path(mod_path)
            # test only one file
            mod_split = mod_abspath.split(sep)
            mod_split.insert(-1, "tests")
            mod_abspath = sep.join(mod_split)
        if not (os.path.exists(mod_abspath)):
            print("Error: no such file or directory: %s" % mod_abspath)
            exit(1)
        modules.append(mod_abspath)
    return modules, extra_args


def nabu_test():
    nabu_folder = get_folder_path()
    args = sys.argv[1:]
    modules_to_test, extra_args = get_modules_to_test(args)
    if len(modules_to_test) == 0:
        modules_to_test = [get_folder_path()]
    pytest_args = extra_args + modules_to_test
    return pytest.main(pytest_args)


if __name__ == "__main__":
    ret = nabu_test()
    exit(ret)



from ...preproc.flatfield import FlatFieldArrays
import numpy as np
from scipy import ndimage as nd
import math


class GriddedAccumulator:
    def __init__(
        self,
        gridded_radios,
        gridded_weights,
        diagnostic_radios,
        diagnostic_weights,
        diagnostic_angles,
        dark=None,
        flat_indexes=None,
        flats=None,
        weights=None,
        double_flat=None,
    ):
        """
        This class creates, for a selected volume slab, a standard set of radios from an helical dataset.

        Parameters
        ==========

        gridded_radios : 3D np.array
           this is the stack of new radios which will be resynthetised, by this class,
           for a selected slab.
           The object is initialised with this array, and this array will accumulate, during subsequent
           calls to  method extract_preprocess_with_flats, the sum of the transformed contributions
           obtained from the arguments of the mentioned method (extract_preprocess_with_flats).
        gridded_weights : 3d np.array
           same shape as gridded_radios, but it will accumulate the weights, during calls to
           extract_preprocess_with_flats
        diagnostic_radios   : 3d np.array,  a stack composed of two radios
           each radio must have the same size as a radio of the gridded_radios argument.
           During the calls to extract_preprocess_with_flats methods,
           the first radio will collect the transformed data for angle=0 ( and the neighbouring ones
           according to angular interpolation coefficients) and this only for the first occurring turn.
           The second radio will be initialised  at the second turn, if any. These array are meant to be used
           to  check the translation step over one turn.
        diagnostic_weights: 3d np.array  a stack composed of two radios
           Same shape as diagnostic_radios. The weigths for diagnostic radios ( will be zero on pixel
           where no data is available, or where the weight is null)
        diagnostic_angles : 1D np.array
           Must have shape==(2,). The two entries will be filled with the angles at which the contributions
           to diagnostic_radios have been summed.
        dark =  None or 2D np.array
           must have the shape of the detector ( generally larger that a radio of gridded_radios)
           If given, the dark will be subtracted from data and flats.
        flat_indexes: None or a list of integers
           the projection index corresponding to the flats
        flats   : None or 3D np.array
           the stack of flats. Each flat must have the shape of the detector  (generally larger that a radio of gridded_radios)
           The flats, if given, are subtracted of the dark, if given, and the result is used to normalise the data.
        weights      : None or 2D np.array
           If not given each data pixel will be considered with unit weight.
           If given it must have the same shape as the detector.
        double_flat  = None or 2D np.array
           If given, the double flat will be applied (division by double_flat)
           Must have  the same shape as the detector.
        """
        self.gridded_radios = gridded_radios
        self.gridded_weights = gridded_weights
        self.diagnostic_radios = diagnostic_radios
        self.diagnostic_weights = diagnostic_weights
        self.diagnostic_angles = diagnostic_angles
        self.dark = dark
        self.flat_indexes = flat_indexes
        self.flat_indexes_reverse_map = dict(
            [(global_index, local_index) for (local_index, global_index) in enumerate(flat_indexes)]
        )
        self.flats = flats
        self.weights = weights
        self.double_flat = double_flat

    def extract_preprocess_with_flats(
        self, subchunk_slice, subchunk_file_indexes, chunk_info, subr_start_end, dtasrc_start_end, data_raw
    ):
        """
        This functions is meant to be called providing, each time, a subset of the data
        which are needed to reconstruct a chunk (to reconstruct a slab).
        When all the necessary data have flown through the subsequent  calls to this method,
        the accumulators are ready.

        Parameters
        ==========
        subchunk_slice: an object of the python class "slice"
          this slice slices the angular domain which corresponds to the useful
          projections  which are useful for the chunk, and whose informations
          are contained in the companion argument "chunk_info"
          Such slicing correspond to the angular subset, for which we are providing
          data_raw
        subchunk_file_indexes: a sequence of integers.
           they correspond to the projection numbers from which the data in data_raw are coming.
           They are used to interpolate between falt fields
        chunk_info: an object returned by the get_chunk_info of the SpanStrategy class
                  this object must have the following members, which relates to the wanted chunk

                  angle_index_span: a pair of integers indicating the start and the end of useful angles
                        in the array of all the scan angle self.projection_angles_deg
                  span_v:  a pair of two integers indicating the start and end of the span relatively  to the lowest value
                        of array self.total_view_heights
                  integer_shift_v: an array, containing for each one of the  useful projections of the span,
                       the integer part of vertical shift to be used in cropping,
                  fract_complement_to_integer_shift_v :
                       the fractional remainder for cropping.
                  z_pix_per_proj: an array, containing for each to be used projection of the span
                       the vertical shift
                  x_pix_per_proj: ....the horizontal shift
                       angles_rad    :    an array, for each useful projection of the chunk the angle in radian
        subr_start_end: a pair of integers
            the start height, and the end height, of the slab for which we are collecting the data.
            The number are given with the same logic as for member span_v of the chunk_info.
            Without the phase margin, when the phase margin is zero, hey would correspond exactly to
            the start and end, vertically, of the reconstructed slices.
        dtasrc_start_end: a pair of integers
            This number are relative to the detector ( they are detector line indexes).
            They indicated, vertically, the detector portion the data_raw data correspond to
        data_raw: np.array 3D
            the data which correspond to a limited detector stripe and a limited angular subset
        """

        # the object below is going to containing some auxiliary variable that are use to reframe the data.
        # This object is used to pass in a compact way such informations to different methods.
        # The informations are extracted from chunk info
        reframing_infos = self._ReframingInfos(
            chunk_info, subchunk_slice, subr_start_end, dtasrc_start_end, subchunk_file_indexes
        )

        # give the proper dimensioning to an auxiliary stack which will contain the reframed data extracted
        # from the treatement of the sub-chunk
        radios_subset = np.zeros(
            [data_raw.shape[0], subr_start_end[1] - subr_start_end[0], data_raw.shape[2]], np.float32
        )

        # ... and in the same way we dimension the container for the  associated reframed weights.
        radios_weights_subset = np.zeros(
            [data_raw.shape[0], subr_start_end[1] - subr_start_end[0], data_raw.shape[2]], np.float32
        )

        # extraction of the data
        self._extract_preprocess_with_flats(data_raw, reframing_infos, chunk_info, radios_subset)

        if self.weights is not None:
            # ... and, if required, extraction of the associated weights
            wdata_read = self.weights.data[reframing_infos.dtasrc_start_z : reframing_infos.dtasrc_end_z]
            self._extract_preprocess_with_flats(
                wdata_read, reframing_infos, chunk_info, radios_weights_subset, it_is_weight=True
            )

        else:
            radios_weights_subset[:] = 1.0

        # and the remaining part is a simple projection over the accumulators,  for
        # the data and for the weights
        my_angles = chunk_info.angles_rad[subchunk_slice]
        n_gridded_angles = self.gridded_radios.shape[0]
        my_i_float = my_angles * (n_gridded_angles / (2 * math.pi))

        tmp_i_rounded = np.floor(my_i_float).astype(np.int32)
        my_epsilon = my_i_float - tmp_i_rounded
        my_i0 = np.mod(tmp_i_rounded, n_gridded_angles)
        my_i1 = np.mod(my_i0 + 1, n_gridded_angles)

        for i0, epsilon, i1, data, weight, original_angle in zip(
            my_i0, my_epsilon, my_i1, radios_subset, radios_weights_subset, chunk_info.angles_rad[subchunk_slice]
        ):
            data_token = data * weight
            self.gridded_radios[i0] += data_token * (1 - epsilon)
            self.gridded_radios[i1] += data_token * epsilon

            self.gridded_weights[i0] += weight * (1 - epsilon)
            self.gridded_weights[i1] += weight * epsilon

            if i0 == 0 or i1 == 0:
                # There is a contribution to the first regridded radio ( the one indexed by 0)
                # We build  two diagnostics for the contributions to this radio.
                # The first for the first pass (i_diag=0)
                # The second for the second pass if any (i_diag=1)

                # To discriminate we introduce
                # An angular margin beyond which we know that a possible contribution
                # is coming from another turn
                safe_angular_margin = 3.14 / 10
                for i_diag in range(2):
                    if original_angle < self.diagnostic_angles[i_diag] + safe_angular_margin:
                        # we are searching for the first contributions ( the one at the lowest angle)
                        # for the two diagnostics. With the constraint that the second is at an higher angle
                        # than the first. So if we are here this means that we have found an occurrence with
                        # lower angle and we discard what we could have previously found.
                        self.diagnostic_radios[i_diag][:] = 0
                        self.diagnostic_weights[i_diag][:] = 0
                        self.diagnostic_angles[i_diag] = original_angle

                    if abs(original_angle - self.diagnostic_angles[i_diag]) < safe_angular_margin:
                        if i0 == 0:
                            factor = 1 - epsilon
                        else:
                            factor = epsilon
                        self.diagnostic_radios[i_diag] += data_token * factor
                        self.diagnostic_weights[i_diag] += weight * factor
                        break

    class _ReframingInfos:
        def __init__(self, chunk_info, subchunk_slice, subr_start_end, dtasrc_start_end, subchunk_file_indexes):
            self.subchunk_file_indexes = subchunk_file_indexes
            my_integer_shifts_v = chunk_info.integer_shift_v[subchunk_slice]
            self.fract_complement_shifts_v = chunk_info.fract_complement_to_integer_shift_v[subchunk_slice]
            self.x_shifts_list = chunk_info.x_pix_per_proj[subchunk_slice]
            subr_start_z, subr_end_z = subr_start_end
            self.subr_start_z_list = subr_start_z - my_integer_shifts_v
            self.subr_end_z_list = subr_end_z - my_integer_shifts_v + 1

            self.dtasrc_start_z, self.dtasrc_end_z = dtasrc_start_end

            floating_start_z = self.subr_start_z_list.min()
            floating_end_z = self.subr_end_z_list.max()

            self.floating_subregion = None, None, floating_start_z, floating_end_z

    def _extract_preprocess_with_flats(self, data_raw, reframing_infos, chunk_info, output, it_is_weight=False):
        if not it_is_weight:
            if self.dark is not None:
                data_raw = data_raw - self.dark[reframing_infos.dtasrc_start_z : reframing_infos.dtasrc_end_z]

            if self.flats is not None:
                for i, idx in enumerate(reframing_infos.subchunk_file_indexes):
                    flat = self._get_flat(idx, slice(reframing_infos.dtasrc_start_z, reframing_infos.dtasrc_end_z))
                    if self.dark is not None:
                        flat = flat - self.dark[reframing_infos.dtasrc_start_z : reframing_infos.dtasrc_end_z]

                    data_raw[i] = data_raw[i] / flat

            if self.double_flat is not None:
                data_raw = data_raw / self.double_flat[reframing_infos.dtasrc_start_z : reframing_infos.dtasrc_end_z]

        if it_is_weight:
            # for the weight, the detector weights, depends on the detector portion,
            # the one corresponding to dtasrc_start_end, and is the same across
            # the subchunk first index ( the projection index)
            take_data_from_this = [data_raw] * len(reframing_infos.subr_start_z_list)
        else:
            take_data_from_this = data_raw

        for data_read, list_subr_start_z, list_subr_end_z, fract_shift, x_shift, data_target in zip(
            take_data_from_this,
            reframing_infos.subr_start_z_list,
            reframing_infos.subr_end_z_list,
            reframing_infos.fract_complement_shifts_v,
            reframing_infos.x_shifts_list,
            output,
        ):
            _fill_in_chunk_by_shift_crop_data(
                data_target,
                data_read,
                fract_shift,
                list_subr_start_z,
                list_subr_end_z,
                reframing_infos.dtasrc_start_z,
                reframing_infos.dtasrc_end_z,
                x_shift=x_shift,
                extension_padding=(not it_is_weight),
            )

    def _get_flat(self, idx, slice_y=slice(None, None), slice_x=slice(None, None), dtype=np.float32):
        prev_next = FlatFieldArrays.get_previous_next_indices(self.flat_indexes, idx)

        if len(prev_next) == 1:  # current index corresponds to an acquired flat
            flat_data = self.flats[self.flat_indexes_reverse_map[prev_next[0]]][slice_y, slice_x]
        else:  # interpolate
            prev_idx, next_idx = prev_next
            flat_data_prev = self.flats[self.flat_indexes_reverse_map[prev_idx]][slice_y, slice_x]
            flat_data_next = self.flats[self.flat_indexes_reverse_map[next_idx]][slice_y, slice_x]
            delta = next_idx - prev_idx
            w1 = 1 - (idx - prev_idx) / delta
            w2 = 1 - (next_idx - idx) / delta
            flat_data = w1 * flat_data_prev + w2 * flat_data_next
        if flat_data.dtype != dtype:
            flat_data = np.ascontiguousarray(flat_data, dtype=dtype)
        return flat_data


def _fill_in_chunk_by_shift_crop_data(
    data_target,
    data_read,
    fract_shift,
    my_subr_start_z,
    my_subr_end_z,
    dtasrc_start_z,
    dtasrc_end_z,
    x_shift=0.0,
    extension_padding=True,
):
    data_read_precisely_shifted = nd.shift(data_read, (-fract_shift, x_shift), order=1, mode="nearest")[:-1]

    target_central_slicer, dtasrc_central_slicer = overlap_logic(
        my_subr_start_z, my_subr_end_z - 1, dtasrc_start_z, dtasrc_end_z - 1
    )

    if None not in [target_central_slicer, dtasrc_central_slicer]:
        data_target[target_central_slicer] = data_read_precisely_shifted[dtasrc_central_slicer]

    target_lower_slicer, target_upper_slicer = padding_logic(
        my_subr_start_z, my_subr_end_z - 1, dtasrc_start_z, dtasrc_end_z - 1
    )

    if extension_padding:
        if target_lower_slicer is not None:
            data_target[target_lower_slicer] = data_read_precisely_shifted[0]
        if target_upper_slicer is not None:
            data_target[target_upper_slicer] = data_read_precisely_shifted[-1]
    else:
        if target_lower_slicer is not None:
            data_target[target_lower_slicer] = 1.0e-6
        if target_upper_slicer is not None:
            data_target[target_upper_slicer] = 1.0e-6


def overlap_logic(subr_start_z, subr_end_z, dtasrc_start_z, dtasrc_end_z):
    """determines the useful lines which can be transferred from the dtasrc_start_z:dtasrc_end_z
    range targeting the range  subr_start_z: subr_end_z ..................

    """

    t_h = subr_end_z - subr_start_z
    s_h = dtasrc_end_z - dtasrc_start_z

    my_start = max(0, dtasrc_start_z - subr_start_z)
    my_end = min(t_h, dtasrc_end_z - subr_start_z)

    if my_start >= my_end:
        return None, None

    target_central_slicer = slice(my_start, my_end)

    my_start = max(0, subr_start_z - dtasrc_start_z)
    my_end = min(s_h, subr_end_z - dtasrc_start_z)

    assert my_start < my_end, "Overlap_logic logic error"

    dtasrc_central_slicer = slice(my_start, my_end)

    return target_central_slicer, dtasrc_central_slicer


def padding_logic(subr_start_z, subr_end_z, dtasrc_start_z, dtasrc_end_z):
    """.......... and the missing ranges which possibly could be obtained by extension padding"""
    t_h = subr_end_z - subr_start_z
    s_h = dtasrc_end_z - dtasrc_start_z

    if dtasrc_start_z <= subr_start_z:
        target_lower_padding = None
    else:
        target_lower_padding = slice(0, dtasrc_start_z - subr_start_z)

    if dtasrc_end_z >= subr_end_z:
        target_upper_padding = None
    else:
        target_upper_padding = slice(dtasrc_end_z - subr_end_z, None)

    return target_lower_padding, target_upper_padding


def get_reconstruction_space(span_info, min_scanwise_z, end_scanwise_z, phase_margin_pix):
    """Utility function which, given the span_info object, creates the auxiliary collection arrays
    and initialises the  my_z_min, my_z_end variable keeping into account the scan direction
    and the min_scanwise_z, end_scanwise_z input arguments
    Parameters
    ==========

    span_info: SpanStrategy

    min_scanwise_z: int
          non negative number, where zero indicates the first feaseable slice doable scanwise.
          Indicates the first (scanwise) requested slice to be reconstructed

    end_scanwise_z: int
          non negative number, where zero indicates the first feaseable slice doable scanwise.
          Indicates the end (scanwise) slice which delimity the to be reconstructed requested slab.

    """

    detector_z_start, detector_z_end = (span_info.get_doable_span()).view_heights_minmax

    if span_info.z_pix_per_proj[-1] > span_info.z_pix_per_proj[0]:
        my_z_min = detector_z_start + min_scanwise_z
        my_z_end = detector_z_start + end_scanwise_z
    else:
        my_z_min = detector_z_end - (end_scanwise_z - 1)
        my_z_end = detector_z_end - (min_scanwise_z + 1)

    # while the raw dataset may have non uniform angular step
    # the regridded dataset will have a constant step.
    # We evaluate here below the number of angles for the
    # regridded dataset, estimating a meaningul angular step representative
    # of the raw data
    my_angle_step = abs(np.diff(span_info.projection_angles_deg).mean())
    n_gridded_angles = int(round(360.0 / my_angle_step))

    radios_h = phase_margin_pix + (my_z_end - my_z_min) + phase_margin_pix

    # the accumulators
    gridded_radios = np.zeros([n_gridded_angles, radios_h, span_info.detector_shape_vh[1]], np.float32)
    gridded_cumulated_weights = np.zeros([n_gridded_angles, radios_h, span_info.detector_shape_vh[1]], np.float32)
    diagnostic_radios = np.zeros((2,) + gridded_radios.shape[1:], np.float32)
    diagnostic_weights = np.zeros((2,) + gridded_radios.shape[1:], np.float32)
    diagnostic_proj_angle = np.zeros([2], "f")

    gridded_angles_rad = np.arange(n_gridded_angles) * 2 * np.pi / n_gridded_angles
    gridded_angles_deg = np.rad2deg(gridded_angles_rad)

    res = type(
        "on_the_fly_class_for_reconstruction_room_in_gridded_accumulator.py",
        (object,),
        {
            "my_z_min": my_z_min,
            "my_z_end": my_z_end,
            "gridded_radios": gridded_radios,
            "gridded_cumulated_weights": gridded_cumulated_weights,
            "diagnostic_radios": diagnostic_radios,
            "diagnostic_weights": diagnostic_weights,
            "diagnostic_proj_angle": diagnostic_proj_angle,
            "gridded_angles_rad": gridded_angles_rad,
            "gridded_angles_deg": gridded_angles_deg,
        },
    )
    return res



import numpy as np
from .flatfield import FlatFieldArrays, load_images_from_dataurl_dict, check_supported


class FlatFieldArraysVariableRegion(FlatFieldArrays):
    _full_shape = True

    def _check_frame_shape(self, frames, frames_type):
        # in helical the flat is the whole one and its shape does not necesseraly match the smaller frames.
        # Therefore no check is done to allow this.
        pass

    def _check_radios_and_indices_congruence(self, radios_indices):
        """At variance with parent class, preprocesing is done with on a fraction of the radios,
        whose length may vary. So we dont enforce here that the lenght is always the same
        """
        pass

    def _normalize_radios(self, radios, sub_indexes, sub_regions_per_radio):
        """
        Apply a flat-field normalization, with the current parameters, to a stack
        of radios.
        The processing is done in-place, meaning that the radios content is overwritten.
        """
        if len(sub_regions_per_radio) != len(sub_indexes):
            message = f""" The length of sub_regions_per_radio,which is {len(sub_regions_per_radio)} , 
            does not correspond to the length of sub_indexes which is {len(sub_indexes)}
            """
            raise ValueError(message)
        do_flats_distortion_correction = self.distortion_correction is not None

        whole_dark = self.get_dark()
        for i, (idx, sub_r) in enumerate(zip(sub_indexes, sub_regions_per_radio)):
            start_x, end_x, start_y, end_y = sub_r
            slice_x = slice(start_x, end_x)
            slice_y = slice(start_y, end_y)

            self.normalize_single_radio(radios[i], idx, dtype=np.float32, slice_y=slice_y, slice_x=slice_x)

        return radios


class FlatFieldDataVariableRegionUrls(FlatFieldArraysVariableRegion):
    def __init__(
        self,
        radios_shape: tuple,
        flats: dict,
        darks: dict,
        radios_indices=None,
        interpolation: str = "linear",
        distortion_correction=None,
        nan_value=1.0,
        radios_srcurrent=None,
        flats_srcurrent=None,
        **chunk_reader_kwargs,
    ):
        flats_arrays_dict = load_images_from_dataurl_dict(flats, **chunk_reader_kwargs)
        darks_arrays_dict = load_images_from_dataurl_dict(darks, **chunk_reader_kwargs)

        _flats_indexes = list(flats_arrays_dict.keys())
        _flats_indexes.sort()

        self.flats_indexes = np.array(_flats_indexes)
        self.flats_stack = np.array([flats_arrays_dict[i] for i in self.flats_indexes], "f")

        flats_arrays_dict = dict([[indx, flat] for indx, flat in zip(self.flats_indexes, self.flats_stack)])

        super().__init__(
            radios_shape,
            flats_arrays_dict,
            darks_arrays_dict,
            radios_indices=radios_indices,
            interpolation=interpolation,
            distortion_correction=distortion_correction,
            nan_value=nan_value,
            radios_srcurrent=radios_srcurrent,
            flats_srcurrent=flats_srcurrent,
        )
        self._sorted_flat_indices = np.array(self._sorted_flat_indices, "i")



import logging
import os
import sys
import numpy as np
import re

from nabu.resources.dataset_analyzer import HDF5DatasetAnalyzer
from nabu.pipeline.estimators import CompositeCOREstimator
from nabu.resources.nxflatfield import update_dataset_info_flats_darks
from .. import version
from .cli_configs import CompositeCorConfig
from .utils import parse_params_values
from ..utils import DictToObj


def main(user_args=None):
    "Application to extract with the composite cor finder the center of rotation for a scan or a series of scans"

    if user_args is None:
        user_args = sys.argv[1:]

    args = DictToObj(
        parse_params_values(
            CompositeCorConfig,
            parser_description=main.__doc__,
            program_version="nabu " + version,
            user_args=user_args,
        )
    )

    if len(os.path.dirname(args.filename_template)) == 0:
        # To make sure that other utility routines can succesfully deal with path within the current directory
        args.filename_template = os.path.join(".", args.filename_template)

    if args.first_stage is not None:
        if args.num_of_stages is None:
            args.num_of_stages = 1
        # if the first_stage parameter has been given then
        # we are using numbers to form the names of the files.
        # The filename must containe a XX..X substring which will be replaced
        pattern = re.compile("[X]+")
        ps = pattern.findall(args.filename_template)
        if len(ps) == 0:
            message = f""" You have specified the "first_stage" parameter, with an integer.
            Therefore the "filename_template" parameter is expected to containe a XX..X subsection
            but none was found in the passed parameter which is {args.filename_template}
            """
            raise ValueError(message)
        ls = list(map(len, ps))
        idx = np.argmax(ls)

        args.filename_template = args.filename_template.replace(ps[idx], "{i_stage:" + "0" + str(ls[idx]) + "d}")

    if args.num_of_stages is None:
        # this way it works also in the simple case where
        # only the filename is provided together with the cor options
        num_of_stages = 1
        first_stage = 0
    else:
        num_of_stages = args.num_of_stages
        first_stage = args.first_stage
    cor_list = []
    for iz in range(first_stage, first_stage + num_of_stages):
        if args.num_of_stages is not None:
            nexus_name = args.filename_template.format(i_stage=iz)
        else:
            nexus_name = args.filename_template

        dataset_info = HDF5DatasetAnalyzer(nexus_name, extra_options={"h5_entry": args.entry_name})
        update_dataset_info_flats_darks(dataset_info, flatfield_mode=1)
        cor_finder = CompositeCOREstimator(
            dataset_info,
            oversampling=args.oversampling,
            theta_interval=args.theta_interval,
            n_subsampling_y=args.n_subsampling_y,
            take_log=True,
            spike_threshold=0.04,
            cor_options=args.cor_options,
        )

        cor_position = cor_finder.find_cor()

        cor_list.append(cor_position)

    cor_list = np.array(cor_list).T

    if args.output_file is not None:
        output_name = args.output_file
    else:
        output_name = os.path.splitext(args.filename_template)[0] + "_cors.txt"

    np.savetxt(
        output_name,
        cor_list,
    )



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "10/05/2022"


from dataclasses import dataclass
import numpy
from tomoscan.identifier import VolumeIdentifier
from tomoscan.identifier import ScanIdentifier
from tomoscan.nexus.paths import nxtomo
from silx.utils.enum import Enum as _Enum
from typing import Optional, Union
from nabu.pipeline.config_validators import (
    integer_validator,
    list_of_shift_validator,
    list_of_tomoscan_identifier,
    optional_directory_location_validator,
    boolean_validator,
    optional_positive_integer_validator,
    output_file_format_validator,
    optional_tuple_of_floats_validator,
    optional_file_name_validator,
)
from nabu.stitching.overlap import OverlapStichingStrategy
from nabu.utils import concatenate_dict, convert_str_to_tuple
from nabu.io.utils import get_output_volume
from tomoscan.factory import Factory


KEY_X_CROSS_CORRELATION_FUNC = "x_cross_correlation_function"

KEY_Y_CROSS_CORRELATION_FUNC = "y_cross_correlation_function"

KEY_CROSS_CORRELATION_SLICE = "do_slice_index_correlation_from_slice"

_DEFAULT_AUTO_REL_SHIFT_PARAMS = (
    f"{KEY_CROSS_CORRELATION_SLICE}=middle;{KEY_X_CROSS_CORRELATION_FUNC}=skimage;{KEY_Y_CROSS_CORRELATION_FUNC}=None"
)

_OUTPUT_SECTION = "output"

_INPUTS_SECTION = "inputs"

_INPUT_DATASETS_FIELD = "input_dataset"

_STITCHING_SECTION = "stitching"

_STITCHING_STRATEGY_FIELD = "stitching_strategy"

_STITCHING_TYPE_FIELD = "type"

_DATA_FILE_FIELD = "location"

_FILE_PREFIX_FIELD = "file_prefix"

_FILE_FORMAT_FIELD = "file_format"

_OVERWRITE_RESULTS_FIELD = "overwrite_results"

_DATA_PATH_FIELD = "data_path"

_X_RELATIVE_SHIFTS_FIELD = "x_relative_shifts_in_px"

_VERTICAL_OVERLAP_FIELD = "vertical_overlap_area_in_px"

_STITCHING_HEIGTH_FIELD = "stitching_height_in_px"

_AUTO_RELATIVE_SHIFT_PARAMS_FIELD = "auto_relative_shifts_params"

_NEXUS_VERSION_FIELD = "nexus_version"

_OUTPUT_DTYPE = "data_type"

_OUTPUT_VOLUME = "output_volume"


def _str_to_dict(my_str):
    """convert a string as key_1=value_2;key_2=value_2 to a dict"""
    res = {}
    for key_value in my_str.split(";"):
        key, value = key_value.split("=")
        res[key] = value
    return res


def _valid_relative_shift_params(my_dict):
    valid_keys = (
        KEY_CROSS_CORRELATION_SLICE,
        KEY_X_CROSS_CORRELATION_FUNC,
        KEY_Y_CROSS_CORRELATION_FUNC,
    )
    for key in my_dict.keys():
        if not key in valid_keys:
            raise KeyError(f"{key} is a unrecognized key")
    return my_dict


def _str_to_int_or_auto(my_str):
    ids = my_str.replace(" ", "").split(",")
    try:
        res = tuple([int(val) if val not in ("auto", "'auto'", '"auto"') else "auto" for val in ids])
    except ValueError:
        raise ValueError(f"Fail to convert {my_str} to a list of int or to 'auto'")
    else:
        if len(res) == 1:
            return res[0]


@dataclass
class _StitchingConfiguration:
    """
    bass class to define stitching configuration
    """

    overlap_height: Union[int, tuple]  # overlap area in pixel between each scan
    stitching_height: Union[None, int, tuple]  # height to take in the overlap to apply stitching
    stitching_strategy: OverlapStichingStrategy
    output_dtype: numpy.dtype
    overwrite_results: bool
    x_shifts: Union[str, tuple]

    def to_dict(self):
        """dump configuration to a dict. Must be serializable because might be dump to HDF5 file"""
        return {
            _STITCHING_SECTION: {
                _VERTICAL_OVERLAP_FIELD: self.overlap_height,
                _STITCHING_HEIGTH_FIELD: self.stitching_height,
                _STITCHING_STRATEGY_FIELD: OverlapStichingStrategy.from_value(self.stitching_strategy).value,
                _X_RELATIVE_SHIFTS_FIELD: self.x_shifts,
            },
            _OUTPUT_SECTION: {
                _OUTPUT_DTYPE: str(self.output_dtype),
                _OVERWRITE_RESULTS_FIELD: self.overwrite_results,
            },
        }


class StitchingType(_Enum):
    Z_PREPROC = "z-preproc"
    Z_POSTPROC = "z-postproc"


@dataclass
class ZStitchingConfiguration(_StitchingConfiguration):
    """
    base class to define z-stitching parameters
    """

    auto_relative_shift_params: dict

    def to_dict(self):
        return concatenate_dict(
            super().to_dict(),
            {
                _STITCHING_SECTION: {
                    _AUTO_RELATIVE_SHIFT_PARAMS_FIELD: ";".join(
                        [f"{key}={value}" for key, value in self.auto_relative_shift_params.items()]
                    ),
                },
            },
        )


@dataclass
class PreProcessedZStitchingConfiguration(ZStitchingConfiguration):
    """
    base class to define z-stitching parameters
    """

    input_scans: tuple  # tuple of ScanBase
    output_file_path: str
    output_data_path: str
    output_nexus_version: Optional[float]

    def to_dict(self):
        return concatenate_dict(
            super().to_dict(),
            {
                _INPUTS_SECTION: {
                    _INPUT_DATASETS_FIELD: [str(scan.get_identifier()) for scan in self.input_scans],
                },
                _OUTPUT_SECTION: {
                    _DATA_FILE_FIELD: self.output_file_path,
                    _DATA_PATH_FIELD: self.output_data_path,
                    _NEXUS_VERSION_FIELD: self.output_nexus_version,
                },
            },
        )

    @staticmethod
    def from_dict(config: dict):
        if not isinstance(config, dict):
            raise TypeError(f"config is expected to be a dict and not {type(config)}")
        inputs_scans_str = config.get(_INPUTS_SECTION, {}).get(_INPUT_DATASETS_FIELD, None)
        if inputs_scans_str in (None, ""):
            input_scans = []
        else:
            input_scans = _get_identifiers(inputs_scans_str)

        output_file_path = config.get(_OUTPUT_SECTION, {}).get(_DATA_FILE_FIELD, None)
        if output_file_path is None:
            raise ValueError("output location not provided")

        nexus_version = config.get(_OUTPUT_SECTION, {}).get(_NEXUS_VERSION_FIELD, None)
        if nexus_version in (None, ""):
            nexus_version = nxtomo.LATEST_VERSION
        else:
            nexus_version = float(nexus_version)

        return PreProcessedZStitchingConfiguration(
            overlap_height=_str_to_int_or_auto(config[_STITCHING_SECTION][_VERTICAL_OVERLAP_FIELD]),
            stitching_height=_str_to_int_or_auto(config[_STITCHING_SECTION].get(_STITCHING_HEIGTH_FIELD, "auto")),
            stitching_strategy=OverlapStichingStrategy.from_value(
                config[_STITCHING_SECTION].get(
                    _STITCHING_STRATEGY_FIELD,
                    OverlapStichingStrategy.COSINUS_WEIGHTS,
                ),
            ),
            x_shifts=_str_to_int_or_auto(config[_STITCHING_SECTION][_X_RELATIVE_SHIFTS_FIELD]),
            auto_relative_shift_params=_valid_relative_shift_params(
                _str_to_dict(
                    config[_STITCHING_SECTION].get(
                        _AUTO_RELATIVE_SHIFT_PARAMS_FIELD,
                        _DEFAULT_AUTO_REL_SHIFT_PARAMS,
                    ),
                ),
            ),
            input_scans=input_scans,
            output_file_path=output_file_path,
            output_data_path=config.get(_OUTPUT_SECTION, {}).get(_DATA_PATH_FIELD, "entry_from_stitchig"),
            overwrite_results=config[_STITCHING_SECTION].get(_OVERWRITE_RESULTS_FIELD, True),
            output_nexus_version=nexus_version,
            output_dtype=config[_OUTPUT_SECTION].get(_OUTPUT_DTYPE, numpy.float32),
        )


@dataclass
class PostProcessedZStitchingConfiguration(ZStitchingConfiguration):
    """
    base class to define z-stitching parameters
    """

    input_volumes: tuple  # tuple of VolumeBase
    output_volume: VolumeIdentifier

    @staticmethod
    def from_dict(config: dict):
        if not isinstance(config, dict):
            raise TypeError(f"config is expected to be a dict and not {type(config)}")
        inputs_volumes_str = config.get(_INPUTS_SECTION, {}).get(_INPUT_DATASETS_FIELD, None)
        if inputs_volumes_str in (None, ""):
            input_volumes = []
        else:
            input_volumes = _get_identifiers(inputs_volumes_str)

        output_volume = get_output_volume(
            location=config.get(_OUTPUT_SECTION, {}).get(_DATA_FILE_FIELD, None),
            file_prefix=config.get(_OUTPUT_SECTION, {}).get(_FILE_PREFIX_FIELD, None),
            file_format=config.get(_OUTPUT_SECTION, {}).get(_FILE_FORMAT_FIELD, "hdf5"),
        )
        # on the next section the one with a default value qre the optionnal one
        return PostProcessedZStitchingConfiguration(
            overlap_height=config[_STITCHING_SECTION][_VERTICAL_OVERLAP_FIELD],
            stitching_height=config[_STITCHING_SECTION].get(_STITCHING_HEIGTH_FIELD, "auto"),
            stitching_strategy=OverlapStichingStrategy.from_value(
                config[_STITCHING_SECTION].get(
                    _STITCHING_STRATEGY_FIELD,
                    OverlapStichingStrategy.COSINUS_WEIGHTS,
                ),
            ),
            x_shifts=config[_STITCHING_SECTION][_X_RELATIVE_SHIFTS_FIELD],
            auto_relative_shift_params=_valid_relative_shift_params(
                _str_to_dict(
                    config[_STITCHING_SECTION].get(
                        _AUTO_RELATIVE_SHIFT_PARAMS_FIELD,
                        _DEFAULT_AUTO_REL_SHIFT_PARAMS,
                    ),
                ),
            ),
            input_volumes=input_volumes,
            output_volume=output_volume,
            overwrite_results=config[_STITCHING_SECTION].get(_OVERWRITE_RESULTS_FIELD, True),
            output_dtype=config[_OUTPUT_SECTION].get(_OUTPUT_DTYPE, numpy.float32),
        )

    def to_dict(self):
        return concatenate_dict(
            super().to_dict(),
            {
                _INPUTS_SECTION: {
                    _INPUT_DATASETS_FIELD: [str(volume.get_identifier()) for volume in self.input_volumes],
                },
                _OUTPUT_SECTION: {
                    _OUTPUT_VOLUME: self.output_volume,
                },
            },
        )


def _get_identifiers(list_identifiers_as_str: str) -> tuple:
    # convert str to a list of str that should represent identifiers
    identifiers_as_str = convert_str_to_tuple(list_identifiers_as_str)
    # convert identifiers as string to IdentifierType instances
    return [Factory.create_tomo_object_from_identifier(identifier_as_str) for identifier_as_str in identifiers_as_str]


def dict_to_config_obj(config: dict):
    if not isinstance(config, dict):
        raise TypeError
    stitching_type = config.get(_STITCHING_SECTION, {}).get(_STITCHING_TYPE_FIELD, None)
    if stitching_type is None:
        raise ValueError("Unagle to find stitching type from config dict")
    else:
        stitching_type = StitchingType.from_value(stitching_type)
        if stitching_type is StitchingType.Z_POSTPROC:
            return PostProcessedZStitchingConfiguration.from_dict(config)
        elif stitching_type is StitchingType.Z_PREPROC:
            return PreProcessedZStitchingConfiguration.from_dict(config)
        else:
            raise NotImplementedError(f"stitching type {stitching_type.value} not handled yet")


def get_default_stitching_config(stitching_type: Optional[Union[StitchingType, str]]) -> tuple:
    """
    Return a default configuration for doing stitching.

    :param stitching_type: if None then return a configuration were use can provide inputs for any
                           of the stitching.
                           Else return config dict dedicated to a particular stitching
    :return: (config, section comments)
    """
    if stitching_type is None:
        return concatenate_dict(z_postproc_stitching_config, z_preproc_stitching_config)

    stitching_type = StitchingType.from_value(stitching_type)
    if stitching_type is StitchingType.Z_POSTPROC:
        return z_postproc_stitching_config
    elif stitching_type is StitchingType.Z_PREPROC:
        return z_preproc_stitching_config
    else:
        raise NotImplementedError


SECTIONS_COMMENTS = {
    _STITCHING_SECTION: "section dedicated to stich parameters\n",
    _OUTPUT_SECTION: "section dedicated to output parameters\n",
    _INPUTS_SECTION: "section dedicated to inputs\n",
}


_stitching_config = {
    _STITCHING_SECTION: {
        _VERTICAL_OVERLAP_FIELD: {
            "default": "auto",
            "help": "Overlap area between two scans in pixel. Can be an int or a list of int. If 'auto' will try to deduce it from the magnification and z_translations value",
            "type": "required",
            "validator": list_of_shift_validator,
        },
        _STITCHING_HEIGTH_FIELD: {
            "default": "auto",
            "help": "Height of the stich to apply on the overlap region. If set to 'auto' then will take the largest one possible (equal overlap height)",
            "type": "advanced",
            "validator": list_of_shift_validator,
        },
        _STITCHING_STRATEGY_FIELD: {
            "default": "cosinus weights",
            "help": f"Policy to apply to compute the overlap area. Must be in {OverlapStichingStrategy.values()}.",
            "type": "required",
        },
    },
    _OUTPUT_SECTION: {
        _OVERWRITE_RESULTS_FIELD: {
            "default": "1",
            "help": "What to do in the case where the output file exists.\nBy default, the output data is never overwritten and the process is interrupted if the file already exists.\nSet this option to 1 if you want to overwrite the output files.",
            "validator": boolean_validator,
            "type": "required",
        },
    },
    _INPUTS_SECTION: {
        _INPUT_DATASETS_FIELD: {
            "default": "",
            "help": f"Dataset to stitch together. Must be volume for {StitchingType.Z_PREPROC.value} and NXtomo for {StitchingType.Z_POSTPROC.value}",
            "type": "required",
        },
    },
}

_z_stitching_config = concatenate_dict(
    _stitching_config,
    {
        _STITCHING_SECTION: {
            _X_RELATIVE_SHIFTS_FIELD: {
                "default": "auto",
                "help": "relative shift between two set of frames or volumes.",
                "type": "required",
                "validator": list_of_shift_validator,
            },
            _AUTO_RELATIVE_SHIFT_PARAMS_FIELD: {
                "default": _DEFAULT_AUTO_REL_SHIFT_PARAMS,
                "help": "options to find shift automatically",
                "type": "advanced",
            },
        },
    },
)


z_preproc_stitching_config = concatenate_dict(
    {
        _STITCHING_SECTION: {
            _STITCHING_TYPE_FIELD: {
                "default": StitchingType.Z_PREPROC.value,
                "help": f"Which type of stitching to do. Must be in {StitchingType.values}",
                "type": "required",
            },
        },
        _OUTPUT_SECTION: {
            _DATA_FILE_FIELD: {
                "default": "",
                "help": "HDF5 file to save the generated NXtomo (.nx extension recommanded).",
                "validator": optional_directory_location_validator,
                "type": "required",
            },
            _FILE_FORMAT_FIELD: {
                "default": "hdf5",
                "help": "Output file format. Available are: hdf5, tiff, jp2, edf",
                "validator": output_file_format_validator,
                "type": "optional",
            },
            _NEXUS_VERSION_FIELD: {
                "default": "",
                "help": "output NXtomo version to use for saving stitched NXtomo. If not provided will take the latest version available.",
                "type": "optional",
            },
        },
    },
    _z_stitching_config,
)


z_postproc_stitching_config = concatenate_dict(
    {
        _STITCHING_SECTION: {
            _STITCHING_TYPE_FIELD: {
                "default": StitchingType.Z_POSTPROC.value,
                "help": f"Which type of stitching to do. Must be in {StitchingType.values}",
                "type": "required",
            },
        },
        _OUTPUT_SECTION: {
            _DATA_FILE_FIELD: {
                "default": "",
                "help": "Directory where the output reconstruction is stored.",
                "validator": optional_directory_location_validator,
                "type": "required",
            },
            _FILE_PREFIX_FIELD: {
                "default": "",
                "help": "File prefix. Optional, by default it is inferred from the scanned dataset.",
                "validator": optional_file_name_validator,
                "type": "optional",
            },
            _FILE_FORMAT_FIELD: {
                "default": "hdf5",
                "help": "Output file format. Available are: hdf5, tiff, jp2, edf",
                "validator": output_file_format_validator,
                "type": "optional",
            },
        },
    },
    _z_stitching_config,
)



"""
nabu.pipeline.estimators: helper classes/functions to estimate parameters of a dataset
(center of rotation, detector tilt, etc).
"""
import inspect
import numpy as np
import scipy.fft  #  pylint: disable=E0611
from silx.io import get_data
from ..preproc.flatfield import FlatFieldDataUrls
from ..estimation.cor import (
    CenterOfRotation,
    CenterOfRotationAdaptiveSearch,
    CenterOfRotationSlidingWindow,
    CenterOfRotationGrowingWindow,
)
from ..estimation.cor_sino import SinoCorInterface
from ..estimation.tilt import CameraTilt
from ..estimation.utils import is_fullturn_scan
from ..resources.logger import LoggerOrPrint
from ..resources.utils import extract_parameters
from ..utils import check_supported, is_int
from .params import tilt_methods
from ..resources.dataset_analyzer import get_0_180_radios
from ..misc.rotation import Rotation
from ..io.reader import ChunkReader
from ..preproc.ccd import Log, CCDFilter
from ..misc import fourier_filters
from .params import cor_methods
from ..io.reader import load_images_from_dataurl_dict


def estimate_cor(method, dataset_info, do_flatfield=True, cor_options_str=None, logger=None):
    logger = LoggerOrPrint(logger)
    cor_options_str = cor_options_str or ""
    check_supported(method, list(cor_methods.keys()), "COR estimation method")
    method = cor_methods[method]

    # Extract CoR parameters from configuration file
    try:
        cor_options = extract_parameters(cor_options_str, sep=";")
    except Exception as exc:
        msg = "Could not extract parameters from cor_options: %s" % (str(exc))
        logger.fatal(msg)
        raise ValueError(msg)

    # Dispatch
    if method in CORFinder.search_methods:
        cor_finder = CORFinder(method, dataset_info, do_flatfield=do_flatfield, cor_options=cor_options, logger=logger)
        estimated_cor = cor_finder.find_cor()
    elif method in SinoCORFinder.search_methods:
        cor_finder = SinoCORFinder(
            method,
            dataset_info,
            slice_idx=cor_options.get("slice_idx", "middle"),
            subsampling=cor_options.get("subsampling", 10),
            do_flatfield=do_flatfield,
            cor_options=cor_options,
            logger=logger,
        )
        estimated_cor = cor_finder.find_cor()
    else:
        composite_options = update_func_kwargs(CompositeCORFinder, cor_options)
        for what in ["cor_options", "logger"]:
            composite_options.pop(what, None)
        cor_finder = CompositeCORFinder(
            dataset_info,
            cor_options=cor_options,
            logger=logger,
            **composite_options,
        )
        estimated_cor = cor_finder.find_cor()
    return estimated_cor


class CORFinderBase:
    """
    A base class for CoR estimators.
    It does common tasks like data reading, flatfield, etc.
    """

    search_methods = {}

    def __init__(self, method, dataset_info, do_flatfield=True, cor_options=None, logger=None):
        """
        Initialize a CORFinder object.

        Parameters
        ----------
        dataset_info: `nabu.resources.dataset_analyzer.DatasetAnalyzer`
            Dataset information structure
        """
        check_supported(method, self.search_methods, "CoR estimation method")
        self.logger = LoggerOrPrint(logger)
        self.dataset_info = dataset_info
        self.do_flatfield = do_flatfield
        self.shape = dataset_info.radio_dims[::-1]
        self._init_cor_finder(method, cor_options)

    def _init_cor_finder(self, method, cor_options):
        self.method = method
        if not isinstance(cor_options, (type(None), dict)):
            raise TypeError(
                f"cor_options is expected to be an optional instance of dict. Get {cor_options} ({type(cor_options)}) instead"
            )
        self.cor_options = cor_options or {}

        cor_class = self.search_methods[method]["class"]
        self.cor_finder = cor_class(logger=self.logger)

        default_lookup_side = "right" if self.dataset_info.is_halftomo else "center"
        lookup_side = self.cor_options.get("side", default_lookup_side)

        self.cor_exec_args = []
        self.cor_exec_args.extend(self.search_methods[method].get("default_args", []))

        # CenterOfRotationSlidingWindow is the only class to have a mandatory argument ("side")
        # TODO - it would be more elegant to have it as a kwarg...
        if len(self.cor_exec_args) > 0 and cor_class is CenterOfRotationSlidingWindow:
            self.cor_exec_args[0] = lookup_side
        #
        self.cor_exec_kwargs = update_func_kwargs(self.cor_finder.find_shift, self.cor_options)


class CORFinder(CORFinderBase):
    """
    Find the Center of Rotation with methods based on two (180-degrees opposed) radios.
    """

    search_methods = {
        "centered": {
            "class": CenterOfRotation,
        },
        "global": {
            "class": CenterOfRotationAdaptiveSearch,
            "default_kwargs": {"low_pass": 1, "high_pass": 20},
        },
        "sliding-window": {
            "class": CenterOfRotationSlidingWindow,
            "default_args": ["center"],
        },
        "growing-window": {
            "class": CenterOfRotationGrowingWindow,
        },
    }

    def __init__(self, method, dataset_info, do_flatfield=True, cor_options=None, logger=None):
        """
        Initialize a CORFinder object.

        Parameters
        ----------
        dataset_info: `nabu.resources.dataset_analyzer.DatasetAnalyzer`
            Dataset information structure
        """
        super().__init__(method, dataset_info, do_flatfield=do_flatfield, cor_options=cor_options, logger=logger)
        self._init_radios()
        self._init_flatfield()
        self._apply_flatfield()
        self._apply_tilt()

    def _init_radios(self):
        self.radios, self._radios_indices = get_0_180_radios(self.dataset_info, return_indices=True)

    def _init_flatfield(self):
        if not (self.do_flatfield):
            return
        self.flatfield = FlatFieldDataUrls(
            self.radios.shape,
            flats=self.dataset_info.flats,
            darks=self.dataset_info.darks,
            radios_indices=self._radios_indices,
            interpolation="linear",
            convert_float=True,
        )

    def _apply_flatfield(self):
        if not (self.do_flatfield):
            return
        self.flatfield.normalize_radios(self.radios)

    def _apply_tilt(self):
        tilt = self.dataset_info.detector_tilt
        if tilt is None:
            return
        self.logger.debug("COREstimator: applying detector tilt correction of %f degrees" % tilt)
        rot = Rotation(self.shape, tilt)
        for i in range(self.radios.shape[0]):
            self.radios[i] = rot.rotate(self.radios[i])

    def find_cor(self):
        """
        Find the center of rotation.

        Returns
        -------
        cor: float
            The estimated center of rotation for the current dataset.
        """
        self.logger.info("Estimating center of rotation")
        self.logger.debug("%s.find_shift(%s)" % (self.cor_finder.__class__.__name__, str(self.cor_exec_kwargs)))
        shift = self.cor_finder.find_shift(
            self.radios[0], np.fliplr(self.radios[1]), *self.cor_exec_args, **self.cor_exec_kwargs
        )
        return self.shape[1] / 2 + shift


# alias
COREstimator = CORFinder


class SinoCORFinder(CORFinderBase):
    """
    A class for finding Center of Rotation based on 360 degrees sinograms.
    This class handles the steps of building the sinogram from raw radios.
    """

    search_methods = {
        "sino-coarse-to-fine": {
            "class": SinoCorInterface,
        },
        "sliding-window": {
            "class": CenterOfRotationSlidingWindow,
            "default_args": ["right"],
        },
        "growing-window": {
            "class": CenterOfRotationGrowingWindow,
        },
    }

    def __init__(
        self, method, dataset_info, slice_idx="middle", subsampling=10, do_flatfield=True, cor_options=None, logger=None
    ):
        """
        Initialize a SinoCORFinder object.

        Other parameters
        ----------------
        The following keys can be set in cor_options.

        slice_idx: int or str
            Which slice index to take for building the sinogram.
            For example slice_idx=0 means that we extract the first line of each projection.
            Value can also be "first", "top", "middle", "last", "bottom".
        subsampling: int, float
            subsampling strategy when building sinograms.
            As building the complete sinogram from raw projections might be tedious, the reading is done with subsampling.
            A positive integer value means the subsampling step (i.e `projections[::subsampling]`).
            A negative integer value means we take -subsampling projections in total.
            A float value indicates the angular step in DEGREES.
        """
        super().__init__(method, dataset_info, do_flatfield=do_flatfield, cor_options=cor_options, logger=logger)
        self._check_360()
        self._set_slice_idx(slice_idx)
        self._set_subsampling(subsampling)
        self._load_raw_sinogram()
        self._flatfield(do_flatfield)
        self._get_sinogram()

    def _check_360(self):
        if self.dataset_info.dataset_scanner.scan_range == 360:
            return
        if not is_fullturn_scan(self.dataset_info.rotation_angles):
            raise ValueError("Sinogram-based Center of Rotation estimation can only be used for 360 degrees scans")

    def _set_slice_idx(self, slice_idx):
        n_z = self.dataset_info.radio_dims[1]
        if isinstance(slice_idx, str):
            str_to_idx = {"top": 0, "first": 0, "middle": n_z // 2, "bottom": n_z - 1, "last": n_z - 1}
            check_supported(slice_idx, str_to_idx.keys(), "slice location")
            slice_idx = str_to_idx[slice_idx]
        self.slice_idx = slice_idx

    def _set_subsampling(self, subsampling):
        projs_idx = sorted(self.dataset_info.projections.keys())
        if is_int(subsampling):
            if subsampling < 0:  # Total number of angles
                n_angles = -subsampling
                indices_float = np.linspace(projs_idx[0], projs_idx[-1], n_angles, endpoint=True)
                self.projs_indices = np.round(indices_float).astype(np.int32).tolist()
            else:  # Subsampling step
                self.projs_indices = projs_idx[::subsampling]
        else:  # Angular step
            raise NotImplementedError()

    def _load_raw_sinogram(self):
        if self.slice_idx is None:
            raise ValueError("Unknow slice index")
        # Subsample projections
        files = {}
        for idx in self.projs_indices:
            files[idx] = self.dataset_info.projections[idx]
        self.files = files
        self.data_reader = ChunkReader(
            self.files,
            sub_region=(None, None, self.slice_idx, self.slice_idx + 1),
            convert_float=True,
        )
        self.data_reader.load_files()
        self._radios = self.data_reader.files_data

    def _flatfield(self, do_flatfield):
        self.do_flatfield = bool(do_flatfield)
        if not self.do_flatfield:
            return
        flatfield = FlatFieldDataUrls(
            self._radios.shape,
            self.dataset_info.flats,
            self.dataset_info.darks,
            radios_indices=self.projs_indices,
            sub_region=(None, None, self.slice_idx, self.slice_idx + 1),
        )
        flatfield.normalize_radios(self._radios)

    def _get_sinogram(self):
        log = Log(self._radios.shape, clip_min=1e-6, clip_max=10.0)
        sinogram = self._radios[:, 0, :].copy()
        log.take_logarithm(sinogram)
        self.sinogram = sinogram

    @staticmethod
    def _split_sinogram(sinogram):
        n_a_2 = sinogram.shape[0] // 2
        img_1, img_2 = sinogram[:n_a_2], sinogram[n_a_2:]
        # "Handle" odd number of projections
        if img_2.shape[0] > img_1.shape[0]:
            img_2 = img_2[:-1, :]
        #
        return img_1, img_2

    def find_cor(self):
        self.logger.info("Estimating center of rotation")
        self.logger.debug("%s.find_shift(%s)" % (self.cor_finder.__class__.__name__, str(self.cor_exec_kwargs)))
        img_1, img_2 = self._split_sinogram(self.sinogram)
        shift = self.cor_finder.find_shift(img_1, img_2, *self.cor_exec_args, **self.cor_exec_kwargs)
        return self.shape[1] / 2 + shift


# alias
SinoCOREstimator = SinoCORFinder


class CompositeCORFinder:
    """
    Class and method to prepare sinogram and calculate COR
    The pseudo sinogram is built with shrinked radios taken every theta_interval degres

    Compared to first writing by Christian Nemoz:
        - gives the same result of the original octave script on the dataset sofar tested
        - The meaning of parameter n_subsampling_y (alias subsampling_y)is now the number of lines which are taken from
          every radio. This is more meaningful in terms of amout of collected information because it
          does not depend on the radio size. Moreover this is what was done in the octave script
        - The spike_threshold has been added with default to 0.04
        - The angular sampling is every 5 degree by default, as it is now the case also in the octave script
        - The finding of the optimal overlap is doing by looping over the possible overlap, according to the overlap.
           After a first testing phase by Elodie we, this part, which is the time consuming part, can be accelerated
           by several order of magnitude without modifing the final result
    """

    _default_cor_options = {"low_pass": 0.4, "high_pass": 10, "side": "center", "near_pos": 0, "near_width": 20}

    def __init__(
        self,
        dataset_info,
        oversampling=4,
        theta_interval=5,
        n_subsampling_y=10,
        take_log=True,
        cor_options=None,
        spike_threshold=0.04,
        logger=None,
    ):
        self.dataset_info = dataset_info
        self.logger = LoggerOrPrint(logger)

        # the algorithm can work for angular ranges larger than 1.2*pi
        # up to an arbitrarily number of turns as it is the case in helical scans
        self.spike_threshold = spike_threshold
        # the following line is necessary for multi-turns scan because the encoders is always
        # in the interval 0-360
        self.unwrapped_rotation_angles = np.unwrap(self.dataset_info.rotation_angles)

        self.angle_min = self.unwrapped_rotation_angles.min()
        self.angle_max = self.unwrapped_rotation_angles.max()

        self.sx, self.sy = self.dataset_info.radio_dims

        if (self.angle_max - self.angle_min) < 1.2 * np.pi:
            useful_span = None
            raise ValueError(
                f"""Sinogram-based Center of Rotation estimation can only be used for scans over more than 180 degrees. 
                                 Your angular span was barely above 180 degrees, it was in fact {((self.angle_max - self.angle_min)/np.pi):.2f} x 180
                                 and it is not considered to be enough by the discriminating condition which requires at least 1.2 half-turns
                              """
            )
        else:
            useful_span = min(np.pi, (self.angle_max - self.angle_min) - np.pi)
            # readapt theta_interval accordingly if the span is smaller than pi
            if useful_span < np.pi:
                theta_interval = theta_interval * useful_span / np.pi

        self._get_cor_options(cor_options)

        self.take_log = take_log
        self.ovs = oversampling
        self.theta_interval = theta_interval

        target_sampling_y = np.round(np.linspace(0, self.sy - 1, n_subsampling_y + 2)).astype(int)[1:-1]

        if self.spike_threshold is not None:
            # take also one line below and on above for each line
            # to provide appropriate margin
            self.sampling_y = np.zeros([3 * len(target_sampling_y)], "i")
            self.sampling_y[0::3] = np.maximum(0, target_sampling_y - 1)
            self.sampling_y[2::3] = np.minimum(self.sy - 1, target_sampling_y + 1)
            self.sampling_y[1::3] = target_sampling_y

            self.ccd_correction = CCDFilter((len(self.sampling_y), self.sx), median_clip_thresh=self.spike_threshold)
        else:
            self.sampling_y = target_sampling_y

        self.nproj = self.dataset_info.n_angles

        my_condition = np.less(self.unwrapped_rotation_angles + np.pi, self.angle_max) * np.less(
            self.unwrapped_rotation_angles, self.angle_min + useful_span
        )

        possibly_probed_angles = self.unwrapped_rotation_angles[my_condition]
        possibly_probed_indices = np.arange(len(self.unwrapped_rotation_angles))[my_condition]

        self.dproj = round(len(possibly_probed_angles) / np.rad2deg(useful_span) * self.theta_interval)

        self.probed_angles = possibly_probed_angles[:: self.dproj]
        self.probed_indices = possibly_probed_indices[:: self.dproj]

        self.absolute_indices = sorted(self.dataset_info.projections.keys())

        my_flats = load_images_from_dataurl_dict(self.dataset_info.flats)

        if my_flats is not None and len(list(my_flats.keys())):
            self.use_flat = True
            self.flatfield = FlatFieldDataUrls(
                (len(self.absolute_indices), self.sy, self.sx),
                self.dataset_info.flats,
                self.dataset_info.darks,
                radios_indices=self.absolute_indices,
                dtype=np.float64,
            )
        else:
            self.use_flat = False

        self.sx, self.sy = self.dataset_info.radio_dims
        self.mlog = Log((1,) + (self.sy, self.sx), clip_min=1e-6, clip_max=10.0)
        self.rcor_abs = round(self.sx / 2.0)
        self.cor_acc = round(self.sx / 2.0)

        self.nprobed = len(self.probed_angles)

        # initialize sinograms and radios arrays
        self.sino = np.zeros([2 * self.nprobed * n_subsampling_y, (self.sx - 1) * self.ovs + 1], "f")
        self._loaded = False

        self.high_pass = self.cor_options["high_pass"]
        img_filter = fourier_filters.get_bandpass_filter(
            (self.sino.shape[0] // 2, self.sino.shape[1]),
            cutoff_lowpass=self.cor_options["low_pass"] * self.ovs,
            cutoff_highpass=self.high_pass * self.ovs,
            use_rfft=False,  # rfft changes the image dimensions lenghts to even if odd
            data_type=np.float64,
        )

        # we are interested in filtering only along the x dimension only
        img_filter[:] = img_filter[0]
        self.img_filter = img_filter

    def _oversample(self, radio):
        """oversampling in the horizontal direction"""
        if self.ovs == 1:
            return radio
        result = np.zeros([radio.shape[0], (radio.shape[1] - 1) * self.ovs + 1], "f")
        result[:, :: self.ovs] = radio

        for i in range(1, self.ovs):
            f = i / self.ovs
            result[:, i :: self.ovs] = (1 - f) * result[:, 0 : -self.ovs : self.ovs] + f * result[
                :, self.ovs :: self.ovs
            ]

        return result

    def _get_cor_options(self, cor_options):
        default_dict = self._default_cor_options.copy()
        if self.dataset_info.is_halftomo:
            default_dict["side"] = "right"

        if cor_options is None or cor_options == "":
            cor_options = {}
        if isinstance(cor_options, str):
            try:
                cor_options = extract_parameters(cor_options, sep=";")
            except Exception as exc:
                msg = "Could not extract parameters from cor_options: %s" % (str(exc))
                self.logger.fatal(msg)
                raise ValueError(msg)
        default_dict.update(cor_options)
        cor_options = default_dict

        self.cor_options = cor_options

    def get_radio(self, image_num):
        #  radio_dataset_idx = self.absolute_indices[image_num]
        radio_dataset_idx = image_num
        data_url = self.dataset_info.projections[radio_dataset_idx]
        radio = get_data(data_url).astype(np.float64)
        if self.use_flat:
            self.flatfield.normalize_single_radio(radio, radio_dataset_idx, dtype=radio.dtype)
        if self.take_log:
            self.mlog.take_logarithm(radio)

        radio = radio[self.sampling_y]
        if self.spike_threshold is not None:
            self.ccd_correction.median_clip_correction(radio, output=radio)
            radio = radio[1::3]
        return radio

    def get_sino(self, reload=False):
        """
        Build sinogram (composite image) from the radio files
        """
        if self._loaded and not reload:
            return self.sino

        sorting_indexes = np.argsort(self.unwrapped_rotation_angles)

        sorted_all_angles = self.unwrapped_rotation_angles[sorting_indexes]
        sorted_angle_indexes = np.arange(len(self.unwrapped_rotation_angles))[sorting_indexes]

        irad = 0
        for prob_a, prob_i in zip(self.probed_angles, self.probed_indices):
            radio1 = self.get_radio(self.absolute_indices[prob_i])
            other_angle = prob_a + np.pi

            insertion_point = np.searchsorted(sorted_all_angles, other_angle)
            if insertion_point > 0 and insertion_point < len(sorted_all_angles):
                other_i_l = sorted_angle_indexes[insertion_point - 1]
                other_i_h = sorted_angle_indexes[insertion_point]
                radio_l = self.get_radio(self.absolute_indices[other_i_l])
                radio_h = self.get_radio(self.absolute_indices[other_i_h])
                f = (other_angle - sorted_all_angles[insertion_point - 1]) / (
                    sorted_all_angles[insertion_point] - sorted_all_angles[insertion_point - 1]
                )
                radio2 = (1 - f) * radio_l + f * radio_h
            else:
                if insertion_point == 0:
                    other_i = sorted_angle_indexes[0]
                elif insertion_point == len(sorted_all_angles):
                    other_i = sorted_angle_indexes[insertion_point - 1]
                radio2 = self.get_radio(self.absolute_indices[other_i])

            self.sino[irad : irad + radio1.shape[0], :] = self._oversample(radio1)
            self.sino[
                irad + self.nprobed * radio1.shape[0] : irad + self.nprobed * radio1.shape[0] + radio1.shape[0], :
            ] = self._oversample(radio2)

            irad = irad + radio1.shape[0]

        self.sino[np.isnan(self.sino)] = 0.0001  # ?
        return self.sino

    def find_cor(self, reload=False):
        self.logger.info("Estimating center of rotation")
        self.logger.debug("%s.find_shift(%s)" % (self.__class__.__name__, self.cor_options))
        self.sinogram = self.get_sino(reload=reload)

        dim_v, dim_h = self.sinogram.shape
        assert dim_v % 2 == 0, " this should not happen "
        dim_v = dim_v // 2

        radio1 = self.sinogram[:dim_v]
        radio2 = self.sinogram[dim_v:]

        orig_sy, orig_ovsd_sx = radio1.shape

        radio1 = scipy.fft.ifftn(
            scipy.fft.fftn(radio1, axes=(-2, -1)) * self.img_filter, axes=(-2, -1)
        ).real  # TODO: convolute only along x
        radio2 = scipy.fft.ifftn(
            scipy.fft.fftn(radio2, axes=(-2, -1)) * self.img_filter, axes=(-2, -1)
        ).real  # TODO: convolute only along x

        tmp_sy, ovsd_sx = radio1.shape
        assert orig_sy == tmp_sy and orig_ovsd_sx == ovsd_sx, "this should not happen"

        if self.cor_options["side"] == "center":
            overlap_min = max(round(ovsd_sx - ovsd_sx / 3), 4)
            overlap_max = min(round(ovsd_sx + ovsd_sx / 3), 2 * ovsd_sx - 4)
        elif self.cor_options["side"] == "right":
            overlap_min = max(4, self.ovs * self.high_pass * 3)
            overlap_max = ovsd_sx
        elif self.cor_options["side"] == "left":
            overlap_min = ovsd_sx
            overlap_max = min(2 * ovsd_sx - 4, 2 * ovsd_sx - self.ovs * self.ovs * self.high_pass * 3)
        elif self.cor_options["side"] == "all":
            overlap_min = max(4, self.ovs * self.high_pass * 3)
            overlap_max = min(2 * ovsd_sx - 4, 2 * ovsd_sx - self.ovs * self.ovs * self.high_pass * 3)

        elif self.cor_options["side"] == "near":
            near_pos = self.cor_options["near_pos"]
            near_width = self.cor_options["near_width"]

            overlap_min = max(4, ovsd_sx - 2 * self.ovs * (near_pos + near_width))
            overlap_max = min(2 * ovsd_sx - 4, ovsd_sx - 2 * self.ovs * (near_pos - near_width))

        else:
            message = f""" The cor options "side" can only have one of the three possible values ["","",""].
            But it has the value "{self.cor_options["side"]}" instead
            """
            raise ValueError(message)

        if overlap_min > overlap_max:
            message = f""" There is no safe search range in find_cor once the margins corresponding to the high_pass filter are discarded.
            Try reducing the low_pass parameter in cor_options
            """
            raise ValueError(message)

        self.logger.info(
            "looking for overlap from min %.2f and max %.2f\n" % (overlap_min / self.ovs, overlap_max / self.ovs)
        )

        best_overlap = overlap_min
        best_value = np.inf
        for z in range(int(overlap_min), int(overlap_max) + 1):
            if z <= ovsd_sx:
                my_z = z
                my_radio1 = radio1
                my_radio2 = radio2
            else:
                my_z = ovsd_sx - (z - ovsd_sx)
                my_radio1 = np.fliplr(radio1)
                my_radio2 = np.fliplr(radio2)

            common_left = np.fliplr(my_radio1[:, ovsd_sx - my_z :])[:, : -(self.ovs * self.high_pass * 2)]
            common_right = my_radio2[:, ovsd_sx - my_z : -(self.ovs * self.high_pass * 2)]
            common = common_right - common_left

            value = np.linalg.norm(common)
            norm_diff2 = value * value

            if common_right.size == 0:
                continue
            norm_right = np.linalg.norm(common_right)
            norm_left = np.linalg.norm(common_left)

            value = norm_diff2 / (norm_right * norm_left)

            min_value = min(best_value, value)

            if min_value == value:
                best_overlap = z
                best_value = min_value
            self.logger.debug(
                "testing an overlap of %.2f pixels, actual best overlap is %.2f pixels over %d\r"
                % (z / self.ovs, best_overlap / self.ovs, ovsd_sx / self.ovs),
            )

        offset = (ovsd_sx - best_overlap) / self.ovs / 2
        cor_abs = (self.sx - 1) / 2 + offset

        return cor_abs


# alias
CompositeCOREstimator = CompositeCORFinder


# Some heavily inelegant things going on here
def get_default_kwargs(func):
    params = inspect.signature(func).parameters
    res = {}
    for param_name, param in params.items():
        if param.default != inspect._empty:
            res[param_name] = param.default
    return res


def update_func_kwargs(func, options):
    res_options = get_default_kwargs(func)
    for option_name, option_val in options.items():
        if option_name in res_options:
            res_options[option_name] = option_val
    return res_options


def get_class_name(class_object):
    return str(class_object).split(".")[-1].strip(">").strip("'").strip('"')


class DetectorTiltEstimator:
    """
    Helper class for detector tilt estimation.
    It automatically chooses the right radios and performs flat-field.
    """

    default_tilt_method = "1d-correlation"
    # Given a tilt angle "a", the maximum deviation caused by the tilt (in pixels) is
    #  N/2 * |sin(a)|  where N is the number of pixels
    # We ignore tilts causing less than 0.25 pixel deviation: N/2*|sin(a)| < tilt_threshold
    tilt_threshold = 0.25

    def __init__(self, dataset_info, do_flatfield=True, logger=None, autotilt_options=None):
        """
        Initialize a detector tilt estimator helper.

        Parameters
        ----------
        dataset_info: `dataset_info` object
            Data structure with the dataset information.
        do_flatfield: bool, optional
            Whether to perform flat field on radios.
        logger: `Logger` object, optional
            Logger object
        autotilt_options: dict, optional
            named arguments to pass to the detector tilt estimator class.
        """
        self._set_params(dataset_info, do_flatfield, logger, autotilt_options)
        self.radios, self.radios_indices = get_0_180_radios(dataset_info, return_indices=True)
        self._init_flatfield()
        self._apply_flatfield()

    def _set_params(self, dataset_info, do_flatfield, logger, autotilt_options):
        self.dataset_info = dataset_info
        self.do_flatfield = bool(do_flatfield)
        self.logger = LoggerOrPrint(logger)
        self._get_autotilt_options(autotilt_options)

    def _init_flatfield(self):
        if not (self.do_flatfield):
            return
        self.flatfield = FlatFieldDataUrls(
            self.radios.shape,
            flats=self.dataset_info.flats,
            darks=self.dataset_info.darks,
            radios_indices=self.radios_indices,
            interpolation="linear",
            convert_float=True,
        )

    def _apply_flatfield(self):
        if not (self.do_flatfield):
            return
        self.flatfield.normalize_radios(self.radios)

    def _get_autotilt_options(self, autotilt_options):
        if autotilt_options is None:
            self.autotilt_options = None
            return
        try:
            autotilt_options = extract_parameters(autotilt_options)
        except Exception as exc:
            msg = "Could not extract parameters from autotilt_options: %s" % (str(exc))
            self.logger.fatal(msg)
            raise ValueError(msg)
        self.autotilt_options = autotilt_options
        if "threshold" in autotilt_options:
            self.tilt_threshold = autotilt_options.pop("threshold")

    def find_tilt(self, tilt_method=None):
        """
        Find the detector tilt.

        Parameters
        ----------
        tilt_method: str, optional
            Which tilt estimation method to use.
        """
        if tilt_method is None:
            tilt_method = self.default_tilt_method
        check_supported(tilt_method, set(tilt_methods.values()), "tilt estimation method")
        self.logger.info("Estimating detector tilt angle")
        autotilt_params = {
            "roi_yxhw": None,
            "median_filt_shape": None,
            "padding_mode": None,
            "peak_fit_radius": 1,
            "high_pass": None,
            "low_pass": None,
        }
        autotilt_params.update(self.autotilt_options or {})
        self.logger.debug("%s(%s)" % ("CameraTilt", str(autotilt_params)))

        tilt_calc = CameraTilt()
        tilt_cor_position, camera_tilt = tilt_calc.compute_angle(
            self.radios[0], np.fliplr(self.radios[1]), method=tilt_method, **autotilt_params
        )
        self.logger.info("Estimated detector tilt angle: %f degrees" % camera_tilt)
        # Ignore too small tilts
        max_deviation = np.max(self.dataset_info.radio_dims) * np.abs(np.sin(np.deg2rad(camera_tilt)))
        if self.dataset_info.is_halftomo:
            max_deviation *= 2
        if max_deviation < self.tilt_threshold:
            self.logger.info(
                "Estimated tilt angle (%.3f degrees) results in %.2f maximum pixels shift, which is below threshold (%.2f pixel). Ignoring the tilt, no correction will be done."
                % (camera_tilt, max_deviation, self.tilt_threshold)
            )
            camera_tilt = None
        return camera_tilt


# alias
TiltFinder = DetectorTiltEstimator
